import asyncio
import time
from datetime import datetime
from typing import Any

import requests
from pydantic import BaseModel, PrivateAttr
from vals.graphql_client import Client
from vals.sdk.exceptions import ValsException
from vals.sdk.inspect_wrapper import InspectWrapper
from vals.sdk.types import (
    ModelCustomOperatorFunctionType,
    ModelFunctionType,
    QuestionAnswerPair,
    RunMetadata,
    RunParameters,
    RunStatus,
    TestResult,
)
from vals.sdk.util import _get_auth_token, be_host, fe_host, get_ariadne_client


class Run(BaseModel):
    id: str
    """Unique identifier for the run."""

    name: str
    """Name of the run."""

    qa_set_id: str | None
    """Unique identifier for the QA set run was created with."""

    test_suite_id: str
    """Unique identifier for the test suite run was created with."""

    test_suite_title: str
    """Title of the test suite run was created with."""

    model: str
    """Model used to perform the run."""

    pass_percentage: float | None
    """Average pass percentage of all tests."""

    pass_rate: float | None
    """Percentage of checks that passed"""

    pass_rate_error: float | None
    """Error margin for pass rate"""

    success_rate: float | None
    """Number of tests where all checks passed"""

    success_rate_error: float | None
    """Error margin for success rate"""

    status: RunStatus
    """Status of the run"""

    archived: bool
    """Whether the run has been archived"""

    text_summary: str
    """Automatically generated summary of common error modes for the run."""

    timestamp: datetime
    """Timestamp of when the run was created."""

    completed_at: datetime | None
    """Timestamp of when the run was completed."""

    parameters: RunParameters
    """Parameters used to create the run."""

    test_results: list[TestResult]
    """List of test results for the run."""

    _client: Client = PrivateAttr(default_factory=get_ariadne_client)

    @staticmethod
    async def _create_from_pull_result(run_id: str, client: Client) -> "Run":
        """Helper method to create a Run instance from a pull_run query result"""

        result = await client.pull_run(run_id)

        offset = 0
        page_size = 200
        test_results = []
        exists_test_results_left_to_pull = True
        while exists_test_results_left_to_pull:
            test_result_query = await client.pull_test_results_with_count(
                run_id=run_id, offset=offset
            )
            test_results.extend(
                [
                    TestResult.from_graphql(test_result)
                    for test_result in test_result_query.test_results_with_count.test_results
                ]
            )
            offset += page_size
            exists_test_results_left_to_pull = (
                len(test_result_query.test_results_with_count.test_results) >= page_size
            )

        # Map maximum_threads to parallelism for backwards compatibility
        parameters_dict = result.run.typed_parameters.model_dump()
        model = parameters_dict.pop("model_under_test", "")

        if "maximum_threads" in parameters_dict:
            parameters_dict["parallelism"] = parameters_dict.pop("maximum_threads")
        parameters = RunParameters(**parameters_dict)

        return Run(
            id=run_id,
            name=result.run.name,
            qa_set_id=result.run.qa_set.id if result.run.qa_set else None,
            model=model,
            pass_percentage=(
                result.run.pass_percentage * 100
                if result.run.pass_percentage is not None
                else None
            ),
            pass_rate=result.run.pass_rate.value if result.run.pass_rate else None,
            pass_rate_error=(
                result.run.pass_rate.error if result.run.pass_rate else None
            ),
            success_rate=(
                result.run.success_rate.value if result.run.success_rate else None
            ),
            success_rate_error=(
                result.run.success_rate.error if result.run.success_rate else None
            ),
            status=RunStatus(result.run.status),
            archived=result.run.archived,
            text_summary=result.run.text_summary,
            timestamp=result.run.timestamp,
            completed_at=result.run.completed_at,
            parameters=parameters,
            test_results=test_results,
            test_suite_title=result.run.test_suite.title,
            test_suite_id=result.run.test_suite.id,
        )

    @classmethod
    async def list_runs(
        cls,
        limit: int = 25,
        offset: int = 0,
        suite_id: str | None = None,
        show_archived: bool = False,
        search: str = "",
    ) -> list["RunMetadata"]:
        """List runs associated with this organization"""
        client = get_ariadne_client()
        result = await client.list_runs(
            limit=limit,
            offset=offset,
            suite_id=suite_id,
            archived=show_archived,
            search=search,
        )
        return [
            RunMetadata.from_graphql(run) for run in result.runs_with_count.run_results
        ]

    @classmethod
    async def from_id(cls, run_id: str) -> "Run":
        """Pull most recent metadata and test results from the vals servers."""
        client = get_ariadne_client()
        return await cls._create_from_pull_result(run_id, client)

    @property
    def url(self) -> str:
        return f"{fe_host()}/results/{self.id}"

    def to_dict(self) -> dict[str, Any]:
        """Converts the run to a dictionary."""
        return self.model_dump(exclude_none=True, exclude_defaults=True, mode="json")

    def to_json_file(self, file_path: str) -> None:
        """Converts the run to a JSON file."""
        with open(file_path, "w") as f:
            f.write(self.to_json_string())

    async def pull(self) -> None:
        """Update this Run instance with latest data from vals servers."""

        updated = await self._create_from_pull_result(self.id, self._client)
        # TODO: There's probably a better way to update the object.
        for field in updated.__fields__:
            setattr(self, field, getattr(updated, field))

    async def get_qa_pairs(
        self, offset: int = 0, limit: int = 1000
    ) -> list[QuestionAnswerPair]:
        """Get all QA pairs for a run."""
        result = await self._client.list_question_answer_pairs(
            qa_set_id=self.qa_set_id, offset=offset, limit=limit
        )
        return [
            QuestionAnswerPair.from_graphql(graphql_qa_pair)
            for graphql_qa_pair in result.question_answer_pairs_with_count.question_answer_pairs
        ]

    async def run_status(self) -> RunStatus:
        """Get the status of a run"""
        result = await self._client.run_status(run_id=self.id)
        status = result.run.status
        self.status = RunStatus(status.lower())
        return self.status

    async def wait_for_run_completion(
        self,
    ) -> RunStatus:
        """
        Block a process until a given run has finished running.

        Returns the status of the run after completion.
        """
        await asyncio.sleep(1)
        status = RunStatus.IN_PROGRESS
        start_time = time.time()
        while status == RunStatus.IN_PROGRESS:
            status = await self.run_status()

            # Sleep longer between polls, the longer the run goes.
            if time.time() - start_time < 60:
                sleep_time = 1
            elif time.time() - start_time < 60 * 10:
                sleep_time = 5
            else:
                sleep_time = 10

            await asyncio.sleep(sleep_time)

        return status

    async def to_csv_string(self) -> str:
        """Same as to_csv, but returns a string instead of writing to a file."""
        response = requests.post(
            url=f"{be_host()}/export_results_to_file/?run_id={self.id}",
            headers={"Authorization": _get_auth_token()},
        )

        if response.status_code != 200:
            raise ValsException("Received Error from Vals Servers: " + response.text)

        return response.text

    async def to_json_string(self) -> str:

        response = requests.post(
            url=f"{be_host()}/export_run_to_json/?run_id={self.id}",
            headers={"Authorization": _get_auth_token()},
        )

        if response.status_code != 200:
            raise ValsException("Received Error from Vals Servers: " + response.text)

        return response.text

    async def to_csv(self, file_path: str) -> None:
        """Get the CSV results of a run, as bytes."""
        with open(file_path, "w") as f:
            f.write(await self.to_csv_string())

    async def retry_failing_tests(self) -> None:
        """Retry all failing tests in a run."""

        await self._client.rerun_tests(run_id=self.id)

    async def resume_run(
        self,
        model: str | ModelFunctionType | list[QuestionAnswerPair] | InspectWrapper,
        wait_for_completion: bool = False,
        upload_concurrency: int | None = None,
        custom_operators: list[ModelCustomOperatorFunctionType] | None = None,
        parallelism: int | None = None,
    ) -> None:
        """Resume a run that was paused.

        This method will:
        1. Check for existing QA pairs that haven't been auto-evaluated
        2. Check for existing completed test results
        3. Run the remaining tests that haven't been processed yet
        """
        if custom_operators is None:
            custom_operators = []

        if parallelism is not None:
            self.parameters.parallelism = parallelism
        # Import Suite here to avoid circular import
        from vals.sdk.suite import Suite

        # Get the test suite
        suite = await Suite.from_id(self.test_suite_id)

        # Create sets of existing test IDs for both QA pairs and completed results
        completed_qa_pairs = {}
        completed_test_results = {}

        qa_pairs = await self.get_qa_pairs(offset=0, limit=len(suite.tests))

        for qa_pair in qa_pairs:
            if len(qa_pair.local_evals) > 0:
                completed_test_results[qa_pair.test_id] = qa_pair
            else:
                completed_qa_pairs[qa_pair.test_id] = qa_pair

        # Filter suite tests to only include those that haven't been processed
        remaining_tests = {
            test._id: test
            for test in suite.tests
            if test._id not in completed_qa_pairs
            and test._id not in completed_test_results
        }

        uploaded_qa_pairs = [
            qa_pair for qa_pair in qa_pairs if qa_pair.test_id in completed_qa_pairs
        ]

        print(f"Remaining model outputs to generate: {len(remaining_tests)}")

        print(
            "Remaining model outputs to evaluate: ",
            len(suite.tests) - len(completed_test_results),
        )

        # Run the remaining tests, including existing QA pairs
        await self._client.update_run_status(
            run_id=self.id, status=RunStatus.IN_PROGRESS.value.upper()
        )
        await suite.run(
            model=model,
            model_name=self.model,
            run_name=self.name,
            wait_for_completion=wait_for_completion,
            parameters=self.parameters,
            upload_concurrency=upload_concurrency,
            custom_operators=custom_operators,
            eval_model_name=self.parameters.eval_model,
            run_id=self.id,
            qa_set_id=self.qa_set_id,
            remaining_tests=list(remaining_tests.values()),
            uploaded_qa_pairs=uploaded_qa_pairs,
        )



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import Any, Optional

from pydantic import Field

from .base_model import BaseModel


class MarkQuestionAnswerSetAsComplete(BaseModel):
    mark_question_answer_set_as_complete: Optional[
        "MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsComplete"
    ] = Field(alias="markQuestionAnswerSetAsComplete")


class MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsComplete(BaseModel):
    question_answer_set: (
        "MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsCompleteQuestionAnswerSet"
    ) = Field(alias="questionAnswerSet")


class MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsCompleteQuestionAnswerSet(
    BaseModel
):
    id: Any


MarkQuestionAnswerSetAsComplete.model_rebuild()
MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsComplete.model_rebuild()



import os
import time

from descope import DescopeClient

VALS_ENV = os.getenv("VALS_ENV", "PROD")

DEFAULT_REGION = "us-east-1"

global_api_key = None
global_in_eu = None
global_auth_dict = {}


def configure_credentials(api_key: str, in_eu: bool = False):
    """
    Configure the Vals API Key to be used with requests.
    This will take precedence over any credentials set in environment variables, or with vals login.

    API key can be generated in the Web App. If you are using the EU platform, make sure to set
    in_eu = True, otherwise leave it as the default.
    """
    global global_api_key, global_in_eu
    global_api_key = api_key.strip()
    global_in_eu = in_eu


def _get_region():
    global global_in_eu
    if global_in_eu is not None:
        return "eu-north-1" if global_in_eu else "us-east-1"

    if "VALS_REGION" in os.environ:
        vals_region = os.environ["VALS_REGION"].lower()
        if vals_region not in ["europe", "us"]:
            raise ValueError(
                f"Invalid region: {vals_region}. Must be 'europe' or 'us'."
            )

        return "eu-north-1" if vals_region == "europe" else "us-east-1"

    return DEFAULT_REGION


def get_descope_client():
    # Needs to be in this file because of circular imports
    region = _get_region()
    if region == "eu-north-1":
        project_id = "P2lXkjgPTaW5f8ZlhBzCpnxeqlpj"
    elif VALS_ENV == "LOCAL" or VALS_ENV == "DEV" or VALS_ENV == "BENCH":
        project_id = "P2ktNOjz5Tgzs9wwS3VpShnCbmik"
    elif VALS_ENV == "PROD":
        project_id = "P2lXkZaPuDqCzGxoxGHseomQi7ac"

    else:
        raise Exception(f"Unrecognized VALS_ENV: {VALS_ENV}")
    return DescopeClient(project_id=project_id, jwt_validation_leeway=30)


def _get_auth_token():
    """
    Get a new session token that can be used in Authorization header.

    Internally, reads the api key from either the VALS_ENV
    variable or a configured value, then uses the Descope SDK to
    exchange the api key for a session token.
    """
    global global_api_key, global_in_eu, global_auth_dict

    # API key was specified with configure_credentials
    if global_api_key is not None:
        api_key = global_api_key

    # API Key is specified in environment
    elif "VALS_API_KEY" in os.environ:
        api_key = os.environ["VALS_API_KEY"]
    else:
        raise Exception(
            "Either the `VALS_API_KEY` environment variable should be set, or the API key should be set with configure_credentials (in vals.sdk.auth.)."
        )

    if (
        "access_expiry" not in global_auth_dict
        or time.time() > global_auth_dict["access_expiry"]
    ):
        descopeClient = get_descope_client()

        response = descopeClient.exchange_access_key(api_key)

        global_auth_dict = {
            **global_auth_dict,
            "access_token": response["sessionToken"]["jwt"],
            "access_expiry": response["sessionToken"]["exp"],
        }

    return global_auth_dict["access_token"]



import asyncio
import base64
import hashlib
import os
import urllib.parse
from collections import defaultdict
from io import BytesIO

import httpx
import requests
from tqdm import tqdm
from vals.graphql_client.client import Client as AriadneClient
from vals.sdk.auth import _get_auth_token, _get_region

VALS_ENV = os.getenv("VALS_ENV")


def read_pdf(file: BytesIO):
    """
    Convenience method to parse PDFs to strings
    """
    try:
        from pypdf import PdfReader
    except ImportError:
        raise Exception(
            "To use read_pdf and read_docx, please run `pip install vals[parsing]`"
        )

    text = ""
    pdf_reader = PdfReader(file)
    num_pages = len(pdf_reader.pages)
    for page_number in range(num_pages):
        page = pdf_reader.pages[page_number]
        text += page.extract_text()
    return text


def read_docx(file: BytesIO):
    """
    Convenience method to parse docx files to strings
    """
    try:
        import pypandoc
    except ImportError:
        raise Exception("To use read_docx, please run `pip install vals[parsing]`")

    output = pypandoc.convert_text(file.read(), to="plain", format="docx")
    return output


def be_host():
    region = _get_region()
    if region == "eu-north-1":
        return "https://europebe.playgroundrl.com"
    if VALS_ENV == "LOCAL":
        return "http://localhost:8000"
    if VALS_ENV == "DEV":
        return "https://devbe.playgroundrl.com"
    if VALS_ENV == "BENCH":
        return "https://benchbe.playgroundrl.com"

    return "https://prodbe.playgroundrl.com"


def fe_host():
    region = _get_region()
    if region == "eu-north-1":
        return "https://eu.platform.vals.ai"
    if VALS_ENV == "LOCAL":
        return "http://localhost:3000"
    if VALS_ENV == "DEV":
        return "https://dev.platform.vals.ai"
    if VALS_ENV == "BENCH":
        return "https://bench.platform.vals.ai"

    return "https://platform.vals.ai"


def get_ariadne_client() -> AriadneClient:
    """
    Use the new codegen-based client
    """

    def append_auth_header(request: httpx.Request):
        request.headers["Authorization"] = _get_auth_token()
        return request

    return AriadneClient(
        url=f"{be_host()}/graphql/",
        http_client=httpx.AsyncClient(auth=append_auth_header, timeout=60),
    )


def md5_hash(file) -> str:
    """Produces an md5 hash of the file."""
    hasher = hashlib.md5()
    while chunk := file.read(8192):  # Read in 8 KB chunks
        hasher.update(chunk)
    hash = hasher.hexdigest()
    file.seek(0)
    return hash


def parse_file_id(file_id: str) -> tuple[str, str, str | None]:
    tokens = file_id.split("/")
    if len(tokens) != 2:
        raise Exception(f"Improperly formatted file_id: {file_id}")

    org, filename_with_hash = tokens

    hash, filename = filename_with_hash.split("-", 1)

    if len(hash) != 32:
        raise Exception(f"Improperly formatted file_id: {file_id}")

    return org, filename, hash


def read_file(file_id: str) -> BytesIO:
    response = requests.post(
        url=f"{be_host()}/download_file/?file_id={file_id}",
        headers={"Authorization": _get_auth_token()},
    )
    return BytesIO(response.content)


async def _download_file_async(file_id: str, client: httpx.AsyncClient):
    """Helper function to download a single file asynchronously"""
    encoded_file_id = urllib.parse.quote(file_id)
    response = await client.get(
        f"{be_host()}/download_files_bulk/?file_ids={encoded_file_id}",
        headers={"Authorization": _get_auth_token()},
    )

    if response.status_code != 200:
        raise Exception(f"Failed to download file {file_id}: {response.text}")

    return response.json()["files"][0]


async def _download_files_chunk_async(file_ids_chunk: list[str]):
    """Download a chunk of files asynchronously"""
    async with httpx.AsyncClient(timeout=60) as client:
        tasks = [_download_file_async(file_id, client) for file_id in file_ids_chunk]
        return await asyncio.gather(*tasks, return_exceptions=True)


async def download_files_bulk(
    file_ids: list[str], download_path: str, max_concurrent_downloads: int = 50
):
    """
    Download multiple files in parallel with a maximum of max_concurrent_downloads downloads at once.

    Args:
        file_ids: List of file IDs to download
        download_path: Path where to save the downloaded files
        max_concurrent_downloads: Maximum number of concurrent downloads (default: 5)
    """

    if len(file_ids) == 0:
        raise Exception("No files to download")

    filename_to_filepath_map = {}

    os.makedirs(download_path, exist_ok=True)

    # Prepare filename groups to handle duplicates
    filename_groups = defaultdict(set)
    for file_id in file_ids:
        id, filename = file_id.split("-", 1)
        hash = id.split("/", 1)[-1]
        filename_groups[filename].add(hash)

    duplicate_files = {
        filename: ids for filename, ids in filename_groups.items() if len(ids) > 1
    }

    # Create directories for duplicate files
    for filename, hashes in duplicate_files.items():
        for hash in hashes:
            download_dir = os.path.join(download_path, hash)
            os.makedirs(download_dir, exist_ok=True)

    # Split file_ids into chunks of max_concurrent_downloads size
    chunks = [
        file_ids[i : i + max_concurrent_downloads]
        for i in range(0, len(file_ids), max_concurrent_downloads)
    ]

    total_files = len(file_ids)
    all_files_data = []

    # Progress bar for downloading
    progress_bar = tqdm(total=total_files, desc="Downloading files", unit="file")

    for chunk in chunks:
        # Use asyncio to download files in parallel
        chunk_results = await _download_files_chunk_async(chunk)

        # Handle any exceptions
        for i, result in enumerate(chunk_results):
            if isinstance(result, Exception):
                print(f"Error downloading {chunk[i]}: {result}")
            else:
                all_files_data.append(result)

            progress_bar.update(1)

    progress_bar.close()

    # Progress bar for saving files
    progress_bar = tqdm(total=len(all_files_data), desc="Saving files", unit="file")

    # Save downloaded files
    for file_data in all_files_data:
        filename = file_data["filename"]
        hash = file_data["hash"]
        content = base64.b64decode(file_data["content"])

        if filename in duplicate_files:
            download_dir = os.path.join(download_path, hash)
            file_path = os.path.join(download_dir, filename)
        else:
            file_path = os.path.join(download_path, filename)

        with open(file_path, "wb") as f:
            f.write(content)

        filename_to_filepath_map[filename] = os.path.abspath(file_path)

        progress_bar.update(1)

    progress_bar.close()

    return filename_to_filepath_map



# Generated by ariadne-codegen
# Source: vals/graphql/

from datetime import datetime
from typing import Any

from pydantic import Field

from .base_model import BaseModel


class GetTestSuiteData(BaseModel):
    test_suite: "GetTestSuiteDataTestSuite" = Field(alias="testSuite")


class GetTestSuiteDataTestSuite(BaseModel):
    description: str
    id: str
    org: str
    title: str
    created: datetime
    global_checks: Any = Field(alias="globalChecks")


GetTestSuiteData.model_rebuild()



class ValsException(Exception):
    """
    An exception returned when there is an error querying the Vals SDK.
    """

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message



import asyncio
import json
from typing import Any, Dict

import click
import requests
from vals.cli.util import display_error_and_exit
from vals.sdk.auth import _get_auth_token
from vals.sdk.util import be_host, get_ariadne_client


@click.group(name="rag")
def rag_group():
    """
    Commands relating to starting rag evaluations.
    """
    pass


async def list_rag_suites():
    client = get_ariadne_client()

    response = await client.get_rag_suites()

    # TODO: Error check
    return response.rag_suites


async def create_rag_suite(data: Dict[str, Any]) -> str:
    client = get_ariadne_client()
    result = await client.create_rag_suite(query=data["query"], file_path=data["id"])
    # Extract the 'id' of the created RagSuite from the result
    suite_id = result.update_rag_suite.rag_suite.id
    return suite_id


async def list_command_async():
    suites = await list_rag_suites()
    if len(suites) == 0:
        click.secho("No rag suites found", fg="red")
        return

    suite_text = "\n".join([f"{i}: {s.id} {s.query}" for i, s in enumerate(suites)])
    click.echo(suite_text)


@click.command(name="list")
def list_command():
    """
    List rag suites associated with this organization
    """
    asyncio.run(list_command_async())


async def upload_command_async(file: str, query: str):
    with open(file, "rb") as f:
        response = requests.post(
            f"{be_host()}/upload_rag_file/",
            files={"file": f},
            headers={"Authorization": _get_auth_token()},
        )
        if response.status_code != 200:
            display_error_and_exit(f"Failed to upload file {file}")

    suite_id = await create_rag_suite(
        {"id": response.json()["file_id"], "query": query}
    )

    click.secho(
        "Successfully Uploaded RAG. ID: {}".format(suite_id),
        fg="green",
    )


@click.command(name="upload")
@click.argument("file", type=click.STRING, required=True)
@click.argument("query", type=click.STRING, required=True)
def upload_command(file: str, query: str):
    """
    Uploads a CSV file to the server for reranking documents.

    Args:
        file (str): The path to the CSV file. The CSV file should
        have a column named "text" containing the documents to be
        reranked as well as a column named id containing the unique
        identifier for each document.

    Returns:
        str: The file ID that can be used to rerank the documents.
    """
    asyncio.run(upload_command_async(file, query))


async def rerank_command_async(
    id: str,
    query: str = "",
    light_ranking: bool = True,
    model: str = "gpt-3",
):
    suites = await list_rag_suites()

    # TODO: Error check
    file_id, original_query = next(
        ((suite.path, suite.query) for suite in suites if suite.id == id),
        None,
    )

    if str(query) == "":
        query = str(original_query)
        click.secho(
            f"Using original query: {query}.",
        )
    else:
        click.secho(
            f"Using custom query: {str(query)}.",
        )

    response = requests.get(
        f"{be_host()}/evaluate_rag_ranking/",
        params={
            "file_id": file_id,
            "query": query,
            "light_ranking": light_ranking,
            "model": model,
        },
        headers={"Authorization": _get_auth_token()},
    )

    if response.status_code != 200:
        display_error_and_exit(f"Failed to run reranking: {file_id}")

    scores = response.json()
    scores["query"] = query

    file_name = f"rag_scores_{id}.json"
    with open(file_name, "w") as f:
        json.dump(scores, f, indent=4)

    click.secho(
        f"Successfully Ran RAG Rankings. Scores saved to {file_name}.",
        fg="green",
    )


@click.command(name="rerank")
@click.argument("id", type=click.STRING, required=True)
@click.option("--query", type=click.STRING, default="")
@click.option("--light-ranking", type=click.STRING, default=True)
@click.option("--model", type=click.STRING, default="gpt-3")
def rerank_command(
    id: str | None = None,
    query: str = "",
    light_ranking: bool = True,
    model: str = "gpt-3",
):
    """
    Reranks the documents in the file based on the provided query.

    Args:
        file_id (str, optional): The ID of the file containing the documents to be reranked.
        query (str, optional): The query to be used for reranking, if not provided the original query will be used.
        light_ranking (bool, optional): Whether to use light ranking. Defaults to True.
        model (str, optional): The model to be used. Defaults to "gpt-3".

    Returns:
        None. Prints a list of scores that match the order of the documents in the file.
    """
    asyncio.run(rerank_command_async(id, query, light_ranking, model))


rag_group.add_command(upload_command)
rag_group.add_command(rerank_command)
rag_group.add_command(list_command, "list")



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import Any, Dict, List, Optional, Union

from .add_batch_tests import AddBatchTests
from .async_base_client import AsyncBaseClient
from .base_model import UNSET, UnsetType
from .batch_add_question_answer_pairs import BatchAddQuestionAnswerPairs
from .create_or_update_test_suite import CreateOrUpdateTestSuite
from .create_question_answer_set import CreateQuestionAnswerSet
from .create_rag_suite import CreateRagSuite
from .delete_test_suite import DeleteTestSuite
from .enums import RunStatus
from .get_active_custom_operators import GetActiveCustomOperators
from .get_operators import GetOperators
from .get_rag_suites import GetRagSuites
from .get_test_data import GetTestData
from .get_test_suite_data import GetTestSuiteData
from .get_test_suites_with_count import GetTestSuitesWithCount
from .input_types import (
    CheckInputType,
    LocalEvalUploadInputType,
    ParameterInputType,
    QuestionAnswerPairInputType,
    TestMutationInfo,
)
from .list_question_answer_pairs import ListQuestionAnswerPairs
from .list_runs import ListRuns
from .mark_question_answer_set_as_complete import MarkQuestionAnswerSetAsComplete
from .pull_run import PullRun
from .pull_test_results_with_count import PullTestResultsWithCount
from .remove_old_tests import RemoveOldTests
from .rerun_tests import RerunTests
from .run_status import RunStatus
from .start_run import StartRun
from .update_global_checks import UpdateGlobalChecks
from .update_run_status import UpdateRunStatus
from .upload_local_evaluation import UploadLocalEvaluation


def gql(q: str) -> str:
    return q


class Client(AsyncBaseClient):
    async def create_question_answer_set(
        self,
        test_suite_id: str,
        question_answer_pairs: List[QuestionAnswerPairInputType],
        parameters: Any,
        model_id: str,
        run_name: Union[Optional[str], UnsetType] = UNSET,
        **kwargs: Any
    ) -> CreateQuestionAnswerSet:
        query = gql(
            """
            mutation CreateQuestionAnswerSet($testSuiteId: String!, $questionAnswerPairs: [QuestionAnswerPairInputType!]!, $parameters: GenericScalar!, $modelId: String!, $runName: String) {
              createQuestionAnswerSet(
                testSuiteId: $testSuiteId
                questionAnswerPairs: $questionAnswerPairs
                parameters: $parameters
                modelId: $modelId
                runName: $runName
                createRun: true
              ) {
                questionAnswerSet {
                  id
                }
                runId
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "testSuiteId": test_suite_id,
            "questionAnswerPairs": question_answer_pairs,
            "parameters": parameters,
            "modelId": model_id,
            "runName": run_name,
        }
        response = await self.execute(
            query=query,
            operation_name="CreateQuestionAnswerSet",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return CreateQuestionAnswerSet.model_validate(data)

    async def batch_add_question_answer_pairs(
        self,
        question_answer_set_id: str,
        question_answer_pairs: List[QuestionAnswerPairInputType],
        **kwargs: Any
    ) -> BatchAddQuestionAnswerPairs:
        query = gql(
            """
            mutation BatchAddQuestionAnswerPairs($questionAnswerSetId: String!, $questionAnswerPairs: [QuestionAnswerPairInputType!]!) {
              batchAddQuestionAnswerPairs(
                questionAnswerSetId: $questionAnswerSetId
                questionAnswerPairs: $questionAnswerPairs
              ) {
                questionAnswerPairs {
                  id
                  inputUnderTest
                  llmOutput
                  fileIds
                  context
                  outputContext
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "questionAnswerSetId": question_answer_set_id,
            "questionAnswerPairs": question_answer_pairs,
        }
        response = await self.execute(
            query=query,
            operation_name="BatchAddQuestionAnswerPairs",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return BatchAddQuestionAnswerPairs.model_validate(data)

    async def mark_question_answer_set_as_complete(
        self, question_answer_set_id: str, **kwargs: Any
    ) -> MarkQuestionAnswerSetAsComplete:
        query = gql(
            """
            mutation MarkQuestionAnswerSetAsComplete($questionAnswerSetId: String!) {
              markQuestionAnswerSetAsComplete(questionAnswerSetId: $questionAnswerSetId) {
                questionAnswerSet {
                  id
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {"questionAnswerSetId": question_answer_set_id}
        response = await self.execute(
            query=query,
            operation_name="MarkQuestionAnswerSetAsComplete",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return MarkQuestionAnswerSetAsComplete.model_validate(data)

    async def list_question_answer_pairs(
        self,
        qa_set_id: str,
        offset: Union[Optional[int], UnsetType] = UNSET,
        limit: Union[Optional[int], UnsetType] = UNSET,
        **kwargs: Any
    ) -> ListQuestionAnswerPairs:
        query = gql(
            """
            query ListQuestionAnswerPairs($qaSetId: String!, $offset: Int, $limit: Int) {
              questionAnswerPairsWithCount(
                qaSetId: $qaSetId
                filterOptions: {offset: $offset, limit: $limit}
              ) {
                questionAnswerPairs {
                  id
                  inputUnderTest
                  llmOutput
                  context
                  outputContext
                  typedMetadata {
                    inTokens
                    outTokens
                    durationSeconds
                  }
                  typedFileIds
                  localEvals {
                    id
                    score
                    feedback
                    createdAt
                  }
                  test {
                    testId
                  }
                }
                count
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "qaSetId": qa_set_id,
            "offset": offset,
            "limit": limit,
        }
        response = await self.execute(
            query=query,
            operation_name="ListQuestionAnswerPairs",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return ListQuestionAnswerPairs.model_validate(data)

    async def start_run(
        self,
        test_suite_id: str,
        typed_parameters: ParameterInputType,
        qa_set_id: Union[Optional[str], UnsetType] = UNSET,
        run_name: Union[Optional[str], UnsetType] = UNSET,
        run_id: Union[Optional[str], UnsetType] = UNSET,
        **kwargs: Any
    ) -> StartRun:
        query = gql(
            """
            mutation startRun($test_suite_id: String!, $typed_parameters: ParameterInputType!, $qa_set_id: String = null, $run_name: String = null, $run_id: String = null) {
              startRun(
                testSuiteId: $test_suite_id
                typedParameters: $typed_parameters
                qaSetId: $qa_set_id
                runName: $run_name
                runId: $run_id
              ) {
                runId
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "test_suite_id": test_suite_id,
            "typed_parameters": typed_parameters,
            "qa_set_id": qa_set_id,
            "run_name": run_name,
            "run_id": run_id,
        }
        response = await self.execute(
            query=query, operation_name="startRun", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return StartRun.model_validate(data)

    async def update_run_status(
        self, run_id: str, status: RunStatus, **kwargs: Any
    ) -> UpdateRunStatus:
        query = gql(
            """
            mutation updateRunStatus($run_id: String!, $status: RunStatus!) {
              updateRunStatus(runId: $run_id, status: $status) {
                run {
                  runId
                  status
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {"run_id": run_id, "status": status}
        response = await self.execute(
            query=query, operation_name="updateRunStatus", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return UpdateRunStatus.model_validate(data)

    async def run_status(self, run_id: str, **kwargs: Any) -> RunStatus:
        query = gql(
            """
            query RunStatus($runId: String!) {
              run(runId: $runId) {
                status
              }
            }
            """
        )
        variables: Dict[str, object] = {"runId": run_id}
        response = await self.execute(
            query=query, operation_name="RunStatus", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return RunStatus.model_validate(data)

    async def pull_run(self, run_id: str, **kwargs: Any) -> PullRun:
        query = gql(
            """
            query PullRun($runId: String!) {
              run(runId: $runId) {
                qaSet {
                  id
                }
                runId
                passPercentage
                status
                textSummary
                timestamp
                completedAt
                archived
                name
                typedParameters {
                  evalModel
                  maximumThreads
                  runGoldenEval
                  runConfidenceEvaluation
                  heavyweightFactor
                  createTextSummary
                  modelUnderTest
                  temperature
                  maxOutputTokens
                  systemPrompt
                  newLineStopOption
                }
                passRate {
                  value
                  error
                }
                successRate {
                  value
                  error
                }
                testSuite {
                  id
                  title
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {"runId": run_id}
        response = await self.execute(
            query=query, operation_name="PullRun", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return PullRun.model_validate(data)

    async def pull_test_results_with_count(
        self,
        run_id: str,
        offset: Union[Optional[int], UnsetType] = UNSET,
        limit: Union[Optional[int], UnsetType] = UNSET,
        **kwargs: Any
    ) -> PullTestResultsWithCount:
        query = gql(
            """
            query PullTestResultsWithCount($runId: String!, $offset: Int, $limit: Int) {
              testResultsWithCount(
                runId: $runId
                filterOptions: {offset: $offset, limit: $limit}
              ) {
                testResults {
                  id
                  llmOutput
                  passPercentage
                  passPercentageWithOptional
                  resultJson
                  qaPair {
                    context
                    outputContext
                    errorMessage
                  }
                  test {
                    testId
                    inputUnderTest
                    context
                  }
                  metadata
                }
                count
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "runId": run_id,
            "offset": offset,
            "limit": limit,
        }
        response = await self.execute(
            query=query,
            operation_name="PullTestResultsWithCount",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return PullTestResultsWithCount.model_validate(data)

    async def list_runs(
        self,
        archived: Union[Optional[bool], UnsetType] = UNSET,
        suite_id: Union[Optional[str], UnsetType] = UNSET,
        limit: Union[Optional[int], UnsetType] = UNSET,
        offset: Union[Optional[int], UnsetType] = UNSET,
        search: Union[Optional[str], UnsetType] = UNSET,
        **kwargs: Any
    ) -> ListRuns:
        query = gql(
            """
            query ListRuns($archived: Boolean, $suiteId: String, $limit: Int, $offset: Int, $search: String) {
              runsWithCount(
                filterOptions: {archived: $archived, suiteId: $suiteId, limit: $limit, offset: $offset, sortBy: STARTED_AT, search: $search}
              ) {
                runResults {
                  runId
                  passPercentage
                  passRate {
                    value
                    error
                  }
                  successRate {
                    value
                    error
                  }
                  name
                  status
                  textSummary
                  timestamp
                  completedAt
                  archived
                  typedParameters {
                    evalModel
                    maximumThreads
                    runGoldenEval
                    runConfidenceEvaluation
                    heavyweightFactor
                    createTextSummary
                    modelUnderTest
                    temperature
                    maxOutputTokens
                    systemPrompt
                    newLineStopOption
                  }
                  testSuite {
                    title
                  }
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "archived": archived,
            "suiteId": suite_id,
            "limit": limit,
            "offset": offset,
            "search": search,
        }
        response = await self.execute(
            query=query, operation_name="ListRuns", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return ListRuns.model_validate(data)

    async def create_or_update_test_suite(
        self, test_suite_id: str, title: str, description: str, **kwargs: Any
    ) -> CreateOrUpdateTestSuite:
        query = gql(
            """
            mutation createOrUpdateTestSuite($testSuiteId: String!, $title: String!, $description: String!) {
              updateTestSuite(
                testSuiteId: $testSuiteId
                title: $title
                description: $description
              ) {
                testSuite {
                  description
                  id
                  org
                  title
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "testSuiteId": test_suite_id,
            "title": title,
            "description": description,
        }
        response = await self.execute(
            query=query,
            operation_name="createOrUpdateTestSuite",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return CreateOrUpdateTestSuite.model_validate(data)

    async def add_batch_tests(
        self, tests: List[TestMutationInfo], create_only: bool, **kwargs: Any
    ) -> AddBatchTests:
        query = gql(
            """
            mutation addBatchTests($tests: [TestMutationInfo!]!, $createOnly: Boolean!) {
              batchUpdateTest(tests: $tests, createOnly: $createOnly) {
                tests {
                  checks
                  testId
                  crossVersionId
                  fileIds
                  inputUnderTest
                  tags
                  context
                  goldenOutput
                  testSuite {
                    id
                  }
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {"tests": tests, "createOnly": create_only}
        response = await self.execute(
            query=query, operation_name="addBatchTests", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return AddBatchTests.model_validate(data)

    async def delete_test_suite(
        self, test_suite_id: str, **kwargs: Any
    ) -> DeleteTestSuite:
        query = gql(
            """
            mutation deleteTestSuite($testSuiteId: String!) {
              deleteSuite(suiteId: $testSuiteId) {
                success
              }
            }
            """
        )
        variables: Dict[str, object] = {"testSuiteId": test_suite_id}
        response = await self.execute(
            query=query, operation_name="deleteTestSuite", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return DeleteTestSuite.model_validate(data)

    async def update_global_checks(
        self, test_suite_id: str, checks: List[CheckInputType], **kwargs: Any
    ) -> UpdateGlobalChecks:
        query = gql(
            """
            mutation updateGlobalChecks($testSuiteId: String!, $checks: [CheckInputType!]!) {
              updateGlobalChecks(testSuiteId: $testSuiteId, checks: $checks) {
                success
              }
            }
            """
        )
        variables: Dict[str, object] = {"testSuiteId": test_suite_id, "checks": checks}
        response = await self.execute(
            query=query,
            operation_name="updateGlobalChecks",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return UpdateGlobalChecks.model_validate(data)

    async def remove_old_tests(
        self, test_suite_id: str, in_use_tests: List[str], **kwargs: Any
    ) -> RemoveOldTests:
        query = gql(
            """
            mutation removeOldTests($testSuiteId: String!, $inUseTests: [String!]!) {
              removeUnusedTests(testSuiteId: $testSuiteId, inUseTests: $inUseTests) {
                success
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "testSuiteId": test_suite_id,
            "inUseTests": in_use_tests,
        }
        response = await self.execute(
            query=query, operation_name="removeOldTests", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return RemoveOldTests.model_validate(data)

    async def create_rag_suite(
        self, query: str, file_path: str, **kwargs: Any
    ) -> CreateRagSuite:
        _query = gql(
            """
            mutation createRagSuite($query: String!, $filePath: String!) {
              updateRagSuite(ragSuiteId: "0", query: $query, filePath: $filePath) {
                ragSuite {
                  id
                  query
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {"query": query, "filePath": file_path}
        response = await self.execute(
            query=_query, operation_name="createRagSuite", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return CreateRagSuite.model_validate(data)

    async def rerun_tests(self, run_id: str, **kwargs: Any) -> RerunTests:
        query = gql(
            """
            mutation RerunTests($runId: String!) {
              rerunFailingTests(runId: $runId) {
                success
              }
            }
            """
        )
        variables: Dict[str, object] = {"runId": run_id}
        response = await self.execute(
            query=query, operation_name="RerunTests", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return RerunTests.model_validate(data)

    async def upload_local_evaluation(
        self,
        question_answer_set_id: str,
        local_evals: List[LocalEvalUploadInputType],
        **kwargs: Any
    ) -> UploadLocalEvaluation:
        query = gql(
            """
            mutation uploadLocalEvaluation($questionAnswerSetId: String!, $localEvals: [LocalEvalUploadInputType!]!) {
              uploadLocalEvaluation(
                questionAnswerSetId: $questionAnswerSetId
                localEvals: $localEvals
              ) {
                uploads {
                  id
                  score
                  feedback
                  createdAt
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "questionAnswerSetId": question_answer_set_id,
            "localEvals": local_evals,
        }
        response = await self.execute(
            query=query,
            operation_name="uploadLocalEvaluation",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return UploadLocalEvaluation.model_validate(data)

    async def get_test_suite_data(
        self, suite_id: str, **kwargs: Any
    ) -> GetTestSuiteData:
        query = gql(
            """
            query getTestSuiteData($suiteId: String!) {
              testSuite(testSuiteId: $suiteId) {
                description
                id
                org
                title
                created
                globalChecks
              }
            }
            """
        )
        variables: Dict[str, object] = {"suiteId": suite_id}
        response = await self.execute(
            query=query,
            operation_name="getTestSuiteData",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return GetTestSuiteData.model_validate(data)

    async def get_test_data(
        self,
        suite_id: str,
        offset: Union[Optional[int], UnsetType] = UNSET,
        limit: Union[Optional[int], UnsetType] = UNSET,
        **kwargs: Any
    ) -> GetTestData:
        query = gql(
            """
            query getTestData($suiteId: String!, $offset: Int, $limit: Int) {
              testsWithCount(
                testSuiteId: $suiteId
                filterOptions: {offset: $offset, limit: $limit}
              ) {
                tests {
                  checks
                  testId
                  crossVersionId
                  fileIds
                  inputUnderTest
                  tags
                  context
                  goldenOutput
                  testSuite {
                    id
                  }
                }
                count
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "suiteId": suite_id,
            "offset": offset,
            "limit": limit,
        }
        response = await self.execute(
            query=query, operation_name="getTestData", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return GetTestData.model_validate(data)

    async def get_test_suites_with_count(
        self,
        offset: Union[Optional[int], UnsetType] = UNSET,
        limit: Union[Optional[int], UnsetType] = UNSET,
        search: Union[Optional[str], UnsetType] = UNSET,
        **kwargs: Any
    ) -> GetTestSuitesWithCount:
        query = gql(
            """
            query getTestSuitesWithCount($offset: Int, $limit: Int, $search: String) {
              testSuitesWithCount(
                filterOptions: {offset: $offset, limit: $limit, search: $search}
              ) {
                testSuites {
                  id
                  title
                  description
                  created
                  creator
                  lastModifiedBy
                  lastModifiedAt
                  folder {
                    id
                    name
                  }
                }
              }
            }
            """
        )
        variables: Dict[str, object] = {
            "offset": offset,
            "limit": limit,
            "search": search,
        }
        response = await self.execute(
            query=query,
            operation_name="getTestSuitesWithCount",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return GetTestSuitesWithCount.model_validate(data)

    async def get_operators(self, **kwargs: Any) -> GetOperators:
        query = gql(
            """
            query getOperators {
              operators {
                nameInDoc
                isUnary
              }
            }
            """
        )
        variables: Dict[str, object] = {}
        response = await self.execute(
            query=query, operation_name="getOperators", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return GetOperators.model_validate(data)

    async def get_active_custom_operators(
        self, offset: int, limit: int, **kwargs: Any
    ) -> GetActiveCustomOperators:
        query = gql(
            """
            query GetActiveCustomOperators($offset: Int!, $limit: Int!) {
              customOperators(
                filterOptions: {archived: false, offset: $offset, limit: $limit}
              ) {
                operators {
                  id
                  name
                  prompt
                  isUnary
                  createdBy
                  createdAt
                  archived
                }
                count
              }
            }
            """
        )
        variables: Dict[str, object] = {"offset": offset, "limit": limit}
        response = await self.execute(
            query=query,
            operation_name="GetActiveCustomOperators",
            variables=variables,
            **kwargs
        )
        data = self.get_data(response)
        return GetActiveCustomOperators.model_validate(data)

    async def get_rag_suites(self, **kwargs: Any) -> GetRagSuites:
        query = gql(
            """
            query getRagSuites {
              ragSuites {
                id
                org
                path
                query
              }
            }
            """
        )
        variables: Dict[str, object] = {}
        response = await self.execute(
            query=query, operation_name="getRagSuites", variables=variables, **kwargs
        )
        data = self.get_data(response)
        return GetRagSuites.model_validate(data)



import asyncio
import json
from io import TextIOWrapper
from typing import Any

import click
from tabulate import tabulate
from vals.cli.util import display_error_and_exit
from vals.sdk.exceptions import ValsException
from vals.sdk.suite import Suite
from vals.sdk.types import RunParameters, RunStatus
from vals.sdk.util import download_files_bulk


@click.group(name="suite")
def suite_group():
    """
    Start, create, or view tests and test suites
    """
    pass


async def create_commmand_async(file: TextIOWrapper):

    suite = await Suite.from_dict(json.loads(file.read()))
    await suite.create()

    click.secho("Successfully created test suite.", fg="green")
    click.secho(f"ID: {suite.id}")
    click.secho(suite.url, bold=True)


@click.command(name="create")
@click.argument("file", type=click.File("r"))
def create_command(file: TextIOWrapper):
    """
    Creates a new test suite based on the json file provided.

    See the documentation for information on the format.
    """
    asyncio.run(create_commmand_async(file))


async def update_command_async(file: TextIOWrapper, suite_id: str):
    try:
        suite = await Suite.from_dict(json.loads(file.read()))
        suite.id = suite_id
        await suite.update()

        click.secho("Successfully updated test suite.", fg="green")
        click.secho(suite.url, bold=True)

    except ValsException as e:
        click.secho(e.message, fg="red")
    except Exception as e:
        click.secho("Suite Update Failed. Error:" + str(e), fg="red")


@click.command(name="update")
@click.argument("file", type=click.File("r"))
@click.argument("suite_id", type=str)
def update_command(file: TextIOWrapper, suite_id: str):
    """
    Update the test and checks of an already existing suite
    """
    asyncio.run(update_command_async(file, suite_id))


async def list_command_async(limit: int, offset: int, search: str):
    suites = await Suite.list_suites(limit=limit, offset=offset - 1, search=search)
    headers = ["#", "Title", "Suite ID", "Last Modified"]
    rows = []
    for i, suite in enumerate(suites, start=offset):
        truncated_title = suite.title
        date_str = suite.last_modified_at.strftime("%Y/%m/%d %H:%M")
        rows.append([i, truncated_title, suite.id, date_str])

    table = tabulate(rows, headers=headers, tablefmt="tsv")
    click.echo(table)


@click.command(name="list")
@click.option("-l", "--limit", type=int, default=25, help="Number of rows to return")
@click.option(
    "-o", "--offset", type=int, default=1, help="Start table at this row (1-indexed)"
)
@click.option("--search", type=str, default="", help="Search for a suite by title")
def list_command(
    limit: int,
    offset: int,
    search: str,
):
    """
    List test suites associated with this organization
    """
    asyncio.run(list_command_async(limit, offset, search))


async def pull_command_async(
    file: TextIOWrapper,
    suite_id: str,
    to_csv: bool,
    to_json: bool,
    download_files: bool,
    download_path: str | None,
    max_concurrent_downloads: int = 50,
):
    if to_csv and to_json:
        display_error_and_exit(
            "Cannot specify both --csv and --json - they are mutually exclusive."
        )

    suite = await Suite.from_id(suite_id)

    if download_files:
        file_ids = []
        for test in suite.tests:
            file_ids.extend(
                [file_id for file_id in test._file_ids if file_id is not None]
            )

        path = suite.title if download_path is None else download_path

        if len(file_ids) != 0:
            click.secho(
                f"This suite has files. Downloading them to directory '{path}'...",
                fg="green",
            )
        filename_to_filepath_map = await download_files_bulk(
            file_ids, path, max_concurrent_downloads=max_concurrent_downloads
        )

    if to_csv:
        file.write(suite.to_csv_string())
    else:
        json_string = suite.to_json_string()
        if download_files:
            parsed_json = json.loads(json_string)
            for i, test in enumerate(parsed_json["tests"]):
                file_paths = []
                for filename in test["files_under_test"]:
                    file_paths.append(filename_to_filepath_map[filename])
                parsed_json["tests"][i]["files_under_test"] = file_paths
            json_string = json.dumps(parsed_json, indent=2)
        file.write(json_string)

    click.secho("Successfully pulled test suite.", fg="green")


@click.command(name="pull")
@click.argument(
    "file",
    type=click.File("w"),
    required=True,
)
@click.argument("suite_id", type=str, required=True)
@click.option("--csv", is_flag=True, help="Output in CSV format")
@click.option("--json", is_flag=True, help="Output in JSON format")
@click.option("--download-files", is_flag=True, help="Download files from the suite")
@click.option(
    "--download-path", type=str, default=None, help="Path to download the files to"
)
@click.option(
    "--max-concurrent-downloads",
    type=int,
    default=50,
    help="Maximum number of concurrent downloads",
)
def pull_command(
    file: TextIOWrapper,
    suite_id: str,
    csv: bool,
    json: bool,
    download_files: bool,
    download_path: str,
    max_concurrent_downloads: int = 50,
):
    """
    Read a suite from the PRL server to a local JSON file.
    """
    asyncio.run(
        pull_command_async(
            file,
            suite_id,
            csv,
            json,
            download_files,
            download_path,
            max_concurrent_downloads,
        )
    )


async def run_command_async(
    suite_id: str,
    model: str,
    run_name: str | None,
    wait_for_completion: bool,
    params: dict[str, Any],
):
    # Remove None values from parameters dictionary
    params = {k: v for k, v in params.items() if v is not None}
    parameters = RunParameters(**params)
    suite = await Suite.from_id(suite_id)

    if wait_for_completion:
        click.echo("Starting run and waiting for it to finish...")

    run = await suite.run(
        model=model,
        run_name=run_name,
        wait_for_completion=wait_for_completion,
        parameters=parameters,
    )

    if wait_for_completion:
        if run.status == RunStatus.SUCCESS:
            click.secho(f"Run has finished successfully", fg="green")
        elif run.status == RunStatus.ERROR:
            click.secho(f"Run has completed with an error", fg="red")
    else:
        click.secho(f"Run has been successfully started", fg="green")

    click.secho(f"Run ID: {run.id}")
    click.secho(run.url, bold=True)


@click.command(name="run")
@click.argument("suite_id", type=str, required=True)
@click.option("--model", type=str, required=True, help="Model to run the tests with")
@click.option(
    "--run-name",
    type=str,
    default=None,
    help="Name of the run",
)
@click.option(
    "--wait-for-completion",
    is_flag=True,
    default=False,
    help="Wait for the run to complete before returning",
)
@click.option(
    "--eval-model", type=str, default=None, help="Model to use for evaluation"
)
@click.option(
    "--parallelism",
    type=int,
    default=None,
    help="Maximum number of concurrent threads",
)
@click.option(
    "--run-golden-eval",
    is_flag=True,
    default=None,
    help="Run evaluation against golden output",
)
@click.option(
    "--run-confidence-evaluation",
    is_flag=True,
    default=None,
    help="Run confidence evaluation",
)
@click.option(
    "--heavyweight-factor",
    type=int,
    default=None,
    help="Heavyweight factor for evaluation",
)
@click.option(
    "--create-text-summary",
    is_flag=True,
    default=None,
    help="Create text summary of results",
)
@click.option(
    "--temperature", type=float, default=None, help="Temperature parameter for model"
)
@click.option(
    "--max-output-tokens", type=int, default=None, help="Maximum tokens in model output"
)
@click.option("--system-prompt", type=str, default=None, help="System prompt for model")
@click.option(
    "--new-line-stop-option", is_flag=True, default=None, help="Stop on new line"
)
def run_command(
    suite_id: str, model: str, run_name: str, wait_for_completion: bool, **params
):
    """
    Run a test suite
    """
    asyncio.run(
        run_command_async(suite_id, model, run_name, wait_for_completion, params)
    )


suite_group.add_command(create_command)
suite_group.add_command(update_command)
suite_group.add_command(list_command)
suite_group.add_command(pull_command)
suite_group.add_command(run_command)



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel


class AddBatchTests(BaseModel):
    batch_update_test: Optional["AddBatchTestsBatchUpdateTest"] = Field(
        alias="batchUpdateTest"
    )


class AddBatchTestsBatchUpdateTest(BaseModel):
    tests: List["AddBatchTestsBatchUpdateTestTests"]


class AddBatchTestsBatchUpdateTestTests(BaseModel):
    checks: Any
    test_id: str = Field(alias="testId")
    cross_version_id: str = Field(alias="crossVersionId")
    file_ids: Any = Field(alias="fileIds")
    input_under_test: str = Field(alias="inputUnderTest")
    tags: Any
    context: Any
    golden_output: str = Field(alias="goldenOutput")
    test_suite: "AddBatchTestsBatchUpdateTestTestsTestSuite" = Field(alias="testSuite")


class AddBatchTestsBatchUpdateTestTestsTestSuite(BaseModel):
    id: str


AddBatchTests.model_rebuild()
AddBatchTestsBatchUpdateTest.model_rebuild()
AddBatchTestsBatchUpdateTestTests.model_rebuild()



# Generated by ariadne-codegen
# Source: http://localhost:8000/graphql/

from enum import Enum


class AppQuestionAnswerSetCreationMethodChoices(str, Enum):
    LIVE_QUERY = "LIVE_QUERY"
    SDK = "SDK"
    UPLOAD = "UPLOAD"


class RunStatus(str, Enum):
    IN_PROGRESS = "IN_PROGRESS"
    ERROR = "ERROR"
    SUCCESS = "SUCCESS"
    RERUNNING = "RERUNNING"


class TestResultReviewStatusEnum(str, Enum):
    PENDING = "PENDING"
    COMPLETED = "COMPLETED"
    ARCHIVED = "ARCHIVED"


class RunReviewStatusEnum(str, Enum):
    PENDING = "PENDING"
    COMPLETED = "COMPLETED"
    ARCHIVED = "ARCHIVED"
    CANCELLED = "CANCELLED"


class AppPairwiseRunReviewWinningRunChoices(str, Enum):
    A = "A"
    B = "B"
    TIE = "TIE"


class AppPairwiseTestResultReviewWinningRunChoices(str, Enum):
    A = "A"
    B = "B"
    TIE = "TIE"


class TierEnum(str, Enum):
    FREE = "FREE"
    STARTUP = "STARTUP"
    ENTERPRISE = "ENTERPRISE"


class TemplateType(str, Enum):
    CATEGORICAL = "CATEGORICAL"
    NUMERICAL = "NUMERICAL"
    FREE_TEXT = "FREE_TEXT"


class RunReviewTableSortField(str, Enum):
    CREATED_AT = "CREATED_AT"
    COMPLETED_TIME = "COMPLETED_TIME"
    STATUS = "STATUS"


class SortOrder(str, Enum):
    ASC = "ASC"
    DESC = "DESC"


class TestResultReviewSortField(str, Enum):
    STARTED_AT = "STARTED_AT"
    STATUS = "STATUS"
    COMPLETED_AT = "COMPLETED_AT"


class RunResultSortField(str, Enum):
    NAME = "NAME"
    STARTED_AT = "STARTED_AT"
    COMPLETED_AT = "COMPLETED_AT"
    PASS_PERCENTAGE = "PASS_PERCENTAGE"


class TestResultCheckErrorEnum(str, Enum):
    ALL = "ALL"
    SOME_CHECKS_FAILED = "SOME_CHECKS_FAILED"
    ALL_CHECKS_FAILED = "ALL_CHECKS_FAILED"
    ALL_CHECKS_PASSED = "ALL_CHECKS_PASSED"
    TESTS_WITH_OUTPUT_ERRORS = "TESTS_WITH_OUTPUT_ERRORS"


class TestSuiteSortField(str, Enum):
    TITLE = "TITLE"
    CREATED = "CREATED"
    LAST_MODIFIED_AT = "LAST_MODIFIED_AT"


class WinningRunEnum(str, Enum):
    A = "A"
    B = "B"
    TIE = "TIE"



# Generated by ariadne-codegen

from .add_batch_tests import (
    AddBatchTests,
    AddBatchTestsBatchUpdateTest,
    AddBatchTestsBatchUpdateTestTests,
    AddBatchTestsBatchUpdateTestTestsTestSuite,
)
from .async_base_client import AsyncBaseClient
from .base_model import BaseModel, Upload
from .batch_add_question_answer_pairs import (
    BatchAddQuestionAnswerPairs,
    BatchAddQuestionAnswerPairsBatchAddQuestionAnswerPairs,
    BatchAddQuestionAnswerPairsBatchAddQuestionAnswerPairsQuestionAnswerPairs,
)
from .client import Client
from .create_or_update_test_suite import (
    CreateOrUpdateTestSuite,
    CreateOrUpdateTestSuiteUpdateTestSuite,
    CreateOrUpdateTestSuiteUpdateTestSuiteTestSuite,
)
from .create_question_answer_set import (
    CreateQuestionAnswerSet,
    CreateQuestionAnswerSetCreateQuestionAnswerSet,
    CreateQuestionAnswerSetCreateQuestionAnswerSetQuestionAnswerSet,
)
from .create_rag_suite import (
    CreateRagSuite,
    CreateRagSuiteUpdateRagSuite,
    CreateRagSuiteUpdateRagSuiteRagSuite,
)
from .delete_test_suite import DeleteTestSuite, DeleteTestSuiteDeleteSuite
from .enums import (
    AppPairwiseRunReviewWinningRunChoices,
    AppPairwiseTestResultReviewWinningRunChoices,
    AppQuestionAnswerSetCreationMethodChoices,
    RunResultSortField,
    RunReviewStatusEnum,
    RunReviewTableSortField,
    RunStatus,
    SortOrder,
    TemplateType,
    TestResultCheckErrorEnum,
    TestResultReviewSortField,
    TestResultReviewStatusEnum,
    TestSuiteSortField,
    TierEnum,
    WinningRunEnum,
)
from .exceptions import (
    GraphQLClientError,
    GraphQLClientGraphQLError,
    GraphQLClientGraphQLMultiError,
    GraphQLClientHttpError,
    GraphQLClientInvalidResponseError,
)
from .get_active_custom_operators import (
    GetActiveCustomOperators,
    GetActiveCustomOperatorsCustomOperators,
    GetActiveCustomOperatorsCustomOperatorsOperators,
)
from .get_operators import GetOperators, GetOperatorsOperators
from .get_rag_suites import GetRagSuites, GetRagSuitesRagSuites
from .get_test_data import (
    GetTestData,
    GetTestDataTestsWithCount,
    GetTestDataTestsWithCountTests,
    GetTestDataTestsWithCountTestsTestSuite,
)
from .get_test_suite_data import GetTestSuiteData, GetTestSuiteDataTestSuite
from .get_test_suites_with_count import (
    GetTestSuitesWithCount,
    GetTestSuitesWithCountTestSuitesWithCount,
    GetTestSuitesWithCountTestSuitesWithCountTestSuites,
    GetTestSuitesWithCountTestSuitesWithCountTestSuitesFolder,
)
from .input_types import (
    CategoricalReviewTemplateInput,
    CheckInputType,
    CheckModifiersInputType,
    ConditionalCheckInputType,
    CustomHumanReviewValueInputType,
    CustomOperatorFilterOptions,
    ExampleInputType,
    FilterOptionsInput,
    FixedOutputInputType,
    FreeTextReviewTemplateInput,
    HumanReviewTemplateFilterOptionsInput,
    LocalEvalUploadInputType,
    MetadataType,
    NumericalReviewTemplateInput,
    ParameterInputType,
    PerCheckTestReviewInputType,
    QuestionAnswerPairInputType,
    QuestionAnswerPairsFilterOptions,
    ResolveApiKeysFilterOptions,
    RunResultFilterOptionsInput,
    RunReviewTableFilterOptionsInput,
    TestFilterOptions,
    TestMutationInfo,
    TestResultFilterOptions,
    TestReviewFilterOptionsInput,
)
from .list_question_answer_pairs import (
    ListQuestionAnswerPairs,
    ListQuestionAnswerPairsQuestionAnswerPairsWithCount,
    ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairs,
    ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsLocalEvals,
    ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTest,
    ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTypedMetadata,
)
from .list_runs import (
    ListRuns,
    ListRunsRunsWithCount,
    ListRunsRunsWithCountRunResults,
    ListRunsRunsWithCountRunResultsPassRate,
    ListRunsRunsWithCountRunResultsSuccessRate,
    ListRunsRunsWithCountRunResultsTestSuite,
    ListRunsRunsWithCountRunResultsTypedParameters,
)
from .mark_question_answer_set_as_complete import (
    MarkQuestionAnswerSetAsComplete,
    MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsComplete,
    MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsCompleteQuestionAnswerSet,
)
from .pull_run import (
    PullRun,
    PullRunRun,
    PullRunRunPassRate,
    PullRunRunQaSet,
    PullRunRunSuccessRate,
    PullRunRunTestSuite,
    PullRunRunTypedParameters,
)
from .pull_test_results_with_count import (
    PullTestResultsWithCount,
    PullTestResultsWithCountTestResultsWithCount,
    PullTestResultsWithCountTestResultsWithCountTestResults,
    PullTestResultsWithCountTestResultsWithCountTestResultsQaPair,
    PullTestResultsWithCountTestResultsWithCountTestResultsTest,
)
from .remove_old_tests import RemoveOldTests, RemoveOldTestsRemoveUnusedTests
from .rerun_tests import RerunTests, RerunTestsRerunFailingTests
from .run_status import RunStatus, RunStatusRun
from .start_run import StartRun, StartRunStartRun
from .update_global_checks import (
    UpdateGlobalChecks,
    UpdateGlobalChecksUpdateGlobalChecks,
)
from .update_run_status import (
    UpdateRunStatus,
    UpdateRunStatusUpdateRunStatus,
    UpdateRunStatusUpdateRunStatusRun,
)
from .upload_local_evaluation import (
    UploadLocalEvaluation,
    UploadLocalEvaluationUploadLocalEvaluation,
    UploadLocalEvaluationUploadLocalEvaluationUploads,
)

__all__ = [
    "AddBatchTests",
    "AddBatchTestsBatchUpdateTest",
    "AddBatchTestsBatchUpdateTestTests",
    "AddBatchTestsBatchUpdateTestTestsTestSuite",
    "AppPairwiseRunReviewWinningRunChoices",
    "AppPairwiseTestResultReviewWinningRunChoices",
    "AppQuestionAnswerSetCreationMethodChoices",
    "AsyncBaseClient",
    "BaseModel",
    "BatchAddQuestionAnswerPairs",
    "BatchAddQuestionAnswerPairsBatchAddQuestionAnswerPairs",
    "BatchAddQuestionAnswerPairsBatchAddQuestionAnswerPairsQuestionAnswerPairs",
    "CategoricalReviewTemplateInput",
    "CheckInputType",
    "CheckModifiersInputType",
    "Client",
    "ConditionalCheckInputType",
    "CreateOrUpdateTestSuite",
    "CreateOrUpdateTestSuiteUpdateTestSuite",
    "CreateOrUpdateTestSuiteUpdateTestSuiteTestSuite",
    "CreateQuestionAnswerSet",
    "CreateQuestionAnswerSetCreateQuestionAnswerSet",
    "CreateQuestionAnswerSetCreateQuestionAnswerSetQuestionAnswerSet",
    "CreateRagSuite",
    "CreateRagSuiteUpdateRagSuite",
    "CreateRagSuiteUpdateRagSuiteRagSuite",
    "CustomHumanReviewValueInputType",
    "CustomOperatorFilterOptions",
    "DeleteTestSuite",
    "DeleteTestSuiteDeleteSuite",
    "ExampleInputType",
    "FilterOptionsInput",
    "FixedOutputInputType",
    "FreeTextReviewTemplateInput",
    "GetActiveCustomOperators",
    "GetActiveCustomOperatorsCustomOperators",
    "GetActiveCustomOperatorsCustomOperatorsOperators",
    "GetOperators",
    "GetOperatorsOperators",
    "GetRagSuites",
    "GetRagSuitesRagSuites",
    "GetTestData",
    "GetTestDataTestsWithCount",
    "GetTestDataTestsWithCountTests",
    "GetTestDataTestsWithCountTestsTestSuite",
    "GetTestSuiteData",
    "GetTestSuiteDataTestSuite",
    "GetTestSuitesWithCount",
    "GetTestSuitesWithCountTestSuitesWithCount",
    "GetTestSuitesWithCountTestSuitesWithCountTestSuites",
    "GetTestSuitesWithCountTestSuitesWithCountTestSuitesFolder",
    "GraphQLClientError",
    "GraphQLClientGraphQLError",
    "GraphQLClientGraphQLMultiError",
    "GraphQLClientHttpError",
    "GraphQLClientInvalidResponseError",
    "HumanReviewTemplateFilterOptionsInput",
    "ListQuestionAnswerPairs",
    "ListQuestionAnswerPairsQuestionAnswerPairsWithCount",
    "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairs",
    "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsLocalEvals",
    "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTest",
    "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTypedMetadata",
    "ListRuns",
    "ListRunsRunsWithCount",
    "ListRunsRunsWithCountRunResults",
    "ListRunsRunsWithCountRunResultsPassRate",
    "ListRunsRunsWithCountRunResultsSuccessRate",
    "ListRunsRunsWithCountRunResultsTestSuite",
    "ListRunsRunsWithCountRunResultsTypedParameters",
    "LocalEvalUploadInputType",
    "MarkQuestionAnswerSetAsComplete",
    "MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsComplete",
    "MarkQuestionAnswerSetAsCompleteMarkQuestionAnswerSetAsCompleteQuestionAnswerSet",
    "MetadataType",
    "NumericalReviewTemplateInput",
    "ParameterInputType",
    "PerCheckTestReviewInputType",
    "PullRun",
    "PullRunRun",
    "PullRunRunPassRate",
    "PullRunRunQaSet",
    "PullRunRunSuccessRate",
    "PullRunRunTestSuite",
    "PullRunRunTypedParameters",
    "PullTestResultsWithCount",
    "PullTestResultsWithCountTestResultsWithCount",
    "PullTestResultsWithCountTestResultsWithCountTestResults",
    "PullTestResultsWithCountTestResultsWithCountTestResultsQaPair",
    "PullTestResultsWithCountTestResultsWithCountTestResultsTest",
    "QuestionAnswerPairInputType",
    "QuestionAnswerPairsFilterOptions",
    "RemoveOldTests",
    "RemoveOldTestsRemoveUnusedTests",
    "RerunTests",
    "RerunTestsRerunFailingTests",
    "ResolveApiKeysFilterOptions",
    "RunResultFilterOptionsInput",
    "RunResultSortField",
    "RunReviewStatusEnum",
    "RunReviewTableFilterOptionsInput",
    "RunReviewTableSortField",
    "RunStatus",
    "RunStatus",
    "RunStatusRun",
    "SortOrder",
    "StartRun",
    "StartRunStartRun",
    "TemplateType",
    "TestFilterOptions",
    "TestMutationInfo",
    "TestResultCheckErrorEnum",
    "TestResultFilterOptions",
    "TestResultReviewSortField",
    "TestResultReviewStatusEnum",
    "TestReviewFilterOptionsInput",
    "TestSuiteSortField",
    "TierEnum",
    "UpdateGlobalChecks",
    "UpdateGlobalChecksUpdateGlobalChecks",
    "UpdateRunStatus",
    "UpdateRunStatusUpdateRunStatus",
    "UpdateRunStatusUpdateRunStatusRun",
    "Upload",
    "UploadLocalEvaluation",
    "UploadLocalEvaluationUploadLocalEvaluation",
    "UploadLocalEvaluationUploadLocalEvaluationUploads",
    "WinningRunEnum",
]



import sys

import click

from .rag import rag_group
from .run import run_group
from .suite import suite_group
from .util import display_error_and_exit


class ExceptionHandlingWrapper(click.Group):
    """Custom class that overrides the default error handling for click"""

    def __call__(self, *args, **kwargs):
        try:
            return super().__call__(*args, **kwargs)
        except Exception as e:
            display_error_and_exit(e)


@click.group(cls=ExceptionHandlingWrapper)
def cli():
    pass


cli.add_command(suite_group)
cli.add_command(run_group)
cli.add_command(rag_group)



import json
import time
from io import BytesIO
from typing import Any, Callable, List, Union

from inspect_ai import Task
from inspect_ai.model import ChatMessageUser, GenerateConfig, ModelOutput, get_model
from inspect_ai.scorer import Score, Scorer, Target
from inspect_ai.solver import Generate, Solver, TaskState
from vals.sdk.types import (
    CustomModelInput,
    CustomModelOutput,
    ModelCustomOperatorFunctionType,
    OperatorInput,
    OperatorOutput,
)


class InspectWrapper:
    """Wrapper to convert between inspect Task components and Vals framework components."""

    model = "vals-sdk"

    def __init__(
        self,
        task: Task,
        model_name: str | None = None,
        config: dict[str, Any] | GenerateConfig | None = None,
        eval_model_name: str | None = None,
    ):
        self.task = task
        self.model_name = model_name
        self.config = config
        self.eval_model_name = eval_model_name

        if model_name is None:
            self.generate = self.get_generate_function(
                "anthropic/claude-3-5-sonnet-20241022", GenerateConfig(temperature=0)
            )
        else:
            self.generate = self.get_generate_function(model_name, config)

    def get_generate_function(
        self, model_name: str, config: dict[str, Any] | GenerateConfig
    ) -> Callable[[TaskState], ModelOutput]:
        if isinstance(config, dict):
            config = GenerateConfig(**config)

        generate = get_model(model_name, config).generate

        async def wrapped_generate(state: str, *args, **kwargs) -> ModelOutput:
            state.output = await generate(state.input)
            return state

        return wrapped_generate

    def custom_model_input_to_task_state(
        self, custom_model_input: CustomModelInput
    ) -> TaskState:
        metadata = (
            custom_model_input.context.get("inspect_context", {})
            .get("task_state", {})
            .get("metadata", {})
        )

        if "uploaded_files" in custom_model_input.context:
            metadata["uploaded_files"] = custom_model_input.context["uploaded_files"]

        if "messages" in custom_model_input.context.get("inspect_context", {}):
            messages = custom_model_input.context["inspect_context"]["messages"]
        else:
            messages = [
                ChatMessageUser(content=custom_model_input.input_under_test),
            ]

        return TaskState(
            model=self.model,
            sample_id=0,
            epoch=0,
            input=messages,
            messages=messages,
            metadata=metadata,
        )

    def operator_input_to_task_state(self, operator_input: OperatorInput) -> TaskState:
        metadata = (
            operator_input.context.get("inspect_context", {})
            .get("task_state", {})
            .get("metadata", {})
        )

        if "uploaded_files" in operator_input.context:
            metadata["uploaded_files"] = operator_input.context["uploaded_files"]

        model_output = ModelOutput()
        model_output.completion = operator_input.model_output

        return TaskState(
            model=self.model,
            sample_id=0,
            epoch=0,
            input=operator_input.input,
            messages=[],
            metadata=metadata,
            output=model_output,
        )

    def operator_input_to_target(self, operator_input: OperatorInput) -> Target:
        return Target(target=operator_input.context["inspect_context"]["target"])

    def get_custom_model(
        self, solver: Solver | None = None, generate: Generate | None = None
    ) -> Callable[[str], Union[str, dict]]:
        """Convert inspect Solver to Vals CustomModel format."""
        solver = solver if solver is not None else self.task.solver
        generate = generate if generate is not None else self.generate

        async def wrapped_custom_model(
            input_under_test: str,
            files: dict[str, BytesIO],
            context: dict[str, Any],
        ) -> CustomModelOutput:
            custom_model_input = CustomModelInput(
                input_under_test=input_under_test,
                files=files,
                context=context,
            )
            # Create TaskState from input
            task_state = self.custom_model_input_to_task_state(custom_model_input)
            # Run solver
            start_time = time.time()
            result_state = await solver(task_state, generate)
            end_time = time.time()

            # Convert to CustomModelOutput format
            return {
                "llm_output": result_state.output.completion,
                "metadata": {
                    "in_tokens": result_state.output.usage.input_tokens,
                    "out_tokens": result_state.output.usage.output_tokens,
                    "duration_seconds": end_time - start_time,
                },
            }

        return wrapped_custom_model

    def get_single_custom_operator(self, scorer: Scorer):
        async def wrapped_custom_operator(
            operator_input: OperatorInput,
        ) -> OperatorOutput:
            task_state = self.operator_input_to_task_state(operator_input)
            target = self.operator_input_to_target(operator_input)

            score: Score = await scorer(task_state, target)

            explanation = score.explanation or ""
            if score.metadata is not None:
                explanation += "\n" + json.dumps(score.metadata)

            output = OperatorOutput(
                name=scorer.__qualname__.split(".")[0],
                score=int(score.value if score.value in [0, 1] else score.value > 0.5),
                explanation=explanation,
            )
            return output

        return wrapped_custom_operator

    def get_custom_operators(
        self, scorer: Scorer | list[Scorer] | None = None
    ) -> List[ModelCustomOperatorFunctionType]:
        """Convert inspect Scorers to Vals CustomOperator format."""
        scorer = scorer if scorer is not None else self.task.scorer
        if not isinstance(scorer, list):
            scorer = [scorer]

        return [self.get_single_custom_operator(r) for r in scorer]



"""
Contains types we define explicitly, as opposed to those that
are auto-generated based on the GraphQL schema.

These are meant to be user-facing.
"""

import datetime
import json
from enum import Enum
from io import BytesIO
from typing import Any, Callable, Literal

from pydantic import BaseModel
from vals.graphql_client.get_test_data import GetTestDataTestsWithCountTests
from vals.graphql_client.get_test_suites_with_count import (
    GetTestSuitesWithCountTestSuitesWithCountTestSuites,
)
from vals.graphql_client.input_types import (
    CheckInputType,
    CheckModifiersInputType,
    LocalEvalUploadInputType,
    MetadataType,
    QuestionAnswerPairInputType,
    TestMutationInfo,
)
from vals.graphql_client.list_question_answer_pairs import (
    ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairs,
)
from vals.graphql_client.list_runs import ListRunsRunsWithCountRunResults
from vals.graphql_client.pull_test_results_with_count import (
    PullTestResultsWithCountTestResultsWithCountTestResults,
)
from vals.sdk.operator_type import OperatorType


class TestSuiteMetadata(BaseModel):
    """
    Data returned about a test suite when we are calling the
    list_test_suites() function - does not include tests, global
    checks, etc.
    """

    id: str
    title: str
    description: str
    created: datetime.datetime
    creator: str
    last_modified_by: str
    last_modified_at: datetime.datetime
    folder_id: str | None
    folder_name: str | None

    # TODO: Often, in this file, we're mapping back and forth between our custom types
    # and the auto-generated types. There is probably a better solution for this.
    @classmethod
    def from_graphql(
        cls, graphql_suite: GetTestSuitesWithCountTestSuitesWithCountTestSuites
    ) -> "TestSuiteMetadata":

        return cls(
            id=graphql_suite.id,
            title=graphql_suite.title,
            description=graphql_suite.description,
            created=graphql_suite.created,
            creator=graphql_suite.creator,
            last_modified_by=graphql_suite.last_modified_by,
            last_modified_at=graphql_suite.last_modified_at,
            folder_id=graphql_suite.folder.id if graphql_suite.folder else None,
            folder_name=graphql_suite.folder.name if graphql_suite.folder else None,
        )


class Example(BaseModel):
    """
    In-context example modifier
    """

    type: Literal["positive", "negative"]
    text: str


class ConditionalCheck(BaseModel):
    """
    Predicate / conditional check modifier.
    """

    operator: OperatorType | str
    criteria: str = ""


class CheckModifiers(BaseModel):
    optional: bool = False
    """Do not include this check towards the final pass percentage"""

    severity: float | None = None
    """Relative weight of this check - see documentation."""

    examples: list[Example] = []
    """In-context examples for the check"""

    extractor: str | None = None
    """Extract certain aspects of the output before the check is evaluated"""

    conditional: ConditionalCheck | None = None
    """Only run this check if another check evaluates to true."""

    category: str | None = None
    """Override the default category of the check (correctness, formatting, etc.)"""

    @classmethod
    def from_graphql(cls, modifiers_dict: dict) -> "CheckModifiers":
        """Internal method to translate from what we receive from GraphQL to the CheckModifiers class."""
        if not modifiers_dict:
            return cls()

        examples = []
        if modifiers_dict.get("examples"):
            examples = [Example(**example) for example in modifiers_dict["examples"]]

        conditional = None
        if modifiers_dict.get("conditional"):
            conditional = ConditionalCheck(**modifiers_dict["conditional"])

        return cls(
            optional=modifiers_dict.get("optional", False),
            severity=modifiers_dict.get("severity"),
            examples=examples,
            extractor=modifiers_dict.get("extractor"),
            conditional=conditional,
            category=modifiers_dict.get("category"),
        )


class Check(BaseModel):
    operator: OperatorType | str
    """Operator - see documentation for the full list."""

    criteria: str = ""
    """Criteria for the check - only required if it is not a unary operator."""

    modifiers: CheckModifiers = CheckModifiers()
    """Optional additional modifiers for the check."""

    @classmethod
    def from_graphql(cls, check_dict: dict) -> "Check":
        """Internal method to translate from what we receive from GraphQL to the Check class displayed to the user."""
        modifiers = CheckModifiers.from_graphql(check_dict.get("modifiers", {}))
        return cls(
            operator=check_dict["operator"],
            criteria=check_dict.get("criteria", ""),
            modifiers=modifiers,
        )

    def to_graphql_input(self) -> dict:
        return CheckInputType(
            operator=self.operator,
            criteria=self.criteria,
            modifiers=CheckModifiersInputType(
                **self.modifiers.model_dump(exclude_none=True)
            ),
        )


class File(BaseModel):
    file_name: str

    file_id: str | None = None

    path: str | None = None

    hash: str | None = None


class Test(BaseModel):
    _id: str = "0"
    """Internal id of the test. 0 signifies it hasn't been created yet."""

    _cross_version_id: str = ""
    """Internal id that stays constant across versions."""

    _test_suite_id: str = ""
    """Maintain internal representation of which Test Suite the test originally belonged to"""

    input_under_test: str
    """Input to the LLM"""

    checks: list[Check]
    """List of checks to apply to the LLM's output"""

    golden_output: str = ""
    """ Expected output of the LLM - can be used instead of or in-conjuction with checks"""

    tags: list[str] = []
    """Tags for the test"""

    context: dict[str, Any] = {}
    """Arbitrary additional context to be used as input for the test."""

    files_under_test: list[File] | list[dict[str, Any]] | list[str] = []
    """Local file paths to upload as part of the test input - i.e. documents, etc."""

    _file_ids: list[str] = []
    """This is the *internal* representation of a file, as stored on the server. You generally should not edit or use these. """

    @classmethod
    def from_graphql_test(cls, graphql_test: GetTestDataTestsWithCountTests) -> "Test":
        """Internal method to translate from what we receive from GraphQL to the Test class displayed to the user."""
        test = cls(
            input_under_test=graphql_test.input_under_test,
            tags=json.loads(graphql_test.tags),
            context=json.loads(graphql_test.context),
            golden_output=graphql_test.golden_output,
            checks=[
                Check.from_graphql(check) for check in json.loads(graphql_test.checks)
            ],
            files_under_test=[
                File(
                    file_name=file.split("-", 1)[-1],
                    file_id=file,
                    hash=file.split("-", 1)[0].split("/", 1)[-1],
                    path=None,
                )
                for file in json.loads(graphql_test.file_ids)
            ],
        )
        test._file_ids = json.loads(graphql_test.file_ids)
        test._id = graphql_test.test_id
        test._cross_version_id = graphql_test.cross_version_id
        test._test_suite_id = graphql_test.test_suite.id
        return test

    def to_test_mutation_info(self, test_suite_id: str) -> TestMutationInfo:
        """Internal method to translate from the Test class to the TestMutationInfo class."""

        file_ids = [file.file_id for file in self.files_under_test]

        return TestMutationInfo(
            test_suite_id=test_suite_id,
            # If we're moving the test to a new test suite, we always need to create it
            test_id=self._id if test_suite_id == self._test_suite_id else "0",
            input_under_test=self.input_under_test,
            checks=[check.to_graphql_input() for check in self.checks],
            tags=self.tags,
            context=json.dumps(self.context),
            golden_output=self.golden_output,
            file_ids=file_ids,
        )


class RunParameters(BaseModel):
    """Parameters for a run."""

    eval_model: str = "gpt-4o"
    """Model to use for the LLM as judge - this is *not* the model being tested."""

    parallelism: int = 10
    """How many tests to run in parallel"""

    run_golden_eval: bool = False
    """Compares the output to the golden ansewr, if provided"""

    run_confidence_evaluation: bool = True
    """ If false, don't produce confidence scores for checks """

    heavyweight_factor: int = 1
    """Run the auto eval multiple times and take the mode of the results"""

    create_text_summary: bool = True
    """If false, will not generate a text summary of the run"""

    temperature: float = 0
    """ Temperature for model being tested"""

    max_output_tokens: int = 512
    """Maximum number of tokens in the output for the model under test"""

    system_prompt: str = ""
    """System prompt for the model under test"""

    new_line_stop_option: bool = False
    """If true, will stop generation at a new line"""

    retry_failed_calls_indefinitely: bool = False
    """ If true, when receiving an error from the model, will retry indefinitely until it receives a success."""


class RunStatus(str, Enum):
    """Status of a run: 'in_progress', 'error', or 'success'."""

    IN_PROGRESS = "in_progress"
    ERROR = "error"
    SUCCESS = "success"
    RERUNNING = "rerunning"


class RunMetadata(BaseModel):
    id: str
    name: str
    pass_percentage: float | None
    pass_rate: float | None
    pass_rate_error: float | None
    success_rate: float | None
    success_rate_error: float | None
    status: RunStatus
    text_summary: str
    timestamp: datetime.datetime
    completed_at: datetime.datetime | None
    archived: bool
    test_suite_title: str
    model: str
    parameters: RunParameters

    @classmethod
    def from_graphql(
        cls, graphql_run: ListRunsRunsWithCountRunResults
    ) -> "RunMetadata":
        return cls(
            id=graphql_run.run_id,
            name=graphql_run.name,
            pass_percentage=(
                graphql_run.pass_percentage if graphql_run.pass_percentage else None
            ),
            pass_rate=graphql_run.pass_rate.value if graphql_run.pass_rate else None,
            pass_rate_error=(
                graphql_run.pass_rate.error if graphql_run.pass_rate else None
            ),
            success_rate=(
                graphql_run.success_rate.value if graphql_run.success_rate else None
            ),
            success_rate_error=(
                graphql_run.success_rate.error if graphql_run.success_rate else None
            ),
            status=RunStatus(graphql_run.status),
            text_summary=graphql_run.text_summary,
            timestamp=graphql_run.timestamp,
            completed_at=graphql_run.completed_at,
            archived=graphql_run.archived,
            test_suite_title=graphql_run.test_suite.title,
            model=graphql_run.typed_parameters.model_under_test,
            parameters=RunParameters(**graphql_run.typed_parameters.model_dump()),
        )


class Metadata(BaseModel):
    in_tokens: int | None = None
    out_tokens: int | None = None
    duration_seconds: float | None = None


class Confidence(Enum):
    """Confidence of a check."""

    LOW = 0
    """Low confidence in the result of the check"""

    CONFIDENCE_NOT_RUN = 0.5
    """Did not run confidence checking on this check"""

    HIGH = 1
    """High confidence in the result of the check"""


class CheckResult(BaseModel):
    """Result of evaluation for a single check."""

    # Same as the input fields.
    operator: OperatorType | str
    criteria: str
    modifiers: CheckModifiers
    is_global: bool

    auto_eval: float
    """Binary pass / fail of the check, 0 for fail, 1 for pass"""

    feedback: str
    """Autogenerated free-text feedback for the check"""

    confidence: Confidence


class TestResult(BaseModel):
    """Result of evaluation for a single test."""

    _id: str
    test_id: str
    input_under_test: str

    context: dict[str, Any]
    """Context pulled from the test"""

    output_context: dict[str, Any]
    """Context produced while producing the output."""

    llm_output: str
    """Output produced by the LLM"""

    metadata: Metadata | None

    pass_percentage: float
    """Percent of passing checks for the test"""

    pass_percentage_with_optional: float
    """Percent of passing checks including optional checks"""

    check_results: list[CheckResult]
    """Results for every check"""

    error_message: str = ""

    @classmethod
    def from_graphql(
        cls,
        graphql_test_result: PullTestResultsWithCountTestResultsWithCountTestResults,
    ) -> "TestResult":
        output_context = {}
        context = {}
        if graphql_test_result.qa_pair:
            output_context = graphql_test_result.qa_pair.output_context
            context = graphql_test_result.qa_pair.context
            if len(context) == 0 and graphql_test_result.test.context is not None:
                context = json.loads(graphql_test_result.test.context)

        return cls(
            _id=graphql_test_result.id,
            test_id=graphql_test_result.test.test_id,
            input_under_test=graphql_test_result.test.input_under_test,
            context=context,
            output_context=output_context,
            llm_output=graphql_test_result.llm_output,
            pass_percentage=graphql_test_result.pass_percentage,
            pass_percentage_with_optional=graphql_test_result.pass_percentage_with_optional,
            error_message=(
                graphql_test_result.qa_pair.error_message
                if graphql_test_result.qa_pair
                else ""
            ),
            check_results=[
                CheckResult(
                    operator=check_result["operator"],
                    criteria=check_result.get("criteria", ""),
                    modifiers=CheckModifiers.from_graphql(
                        check_result.get("modifiers", {})
                    ),
                    is_global=check_result.get("is_global", False),
                    auto_eval=check_result.get("auto_eval", 0),
                    feedback=check_result.get("feedback", ""),
                    confidence=Confidence(
                        check_result.get("eval_cont", 0.5)
                        if check_result.get("eval_cont", 0.5) in [0, 0.5, 1]
                        else 0.5
                    ),
                )
                for check_result in json.loads(graphql_test_result.result_json)
            ],
            metadata=(
                Metadata(**json.loads(graphql_test_result.metadata))
                if graphql_test_result.metadata
                else None
            ),
        )


class QuestionAnswerPair(BaseModel):
    id: str | None = None
    input_under_test: str
    llm_output: str
    file_ids: list[str] | None = None
    context: dict[str, Any] = {}
    output_context: dict[str, Any] = {}
    metadata: Metadata | None = None
    test_id: str | None = None
    local_evals: list[LocalEvalUploadInputType] | None = None

    @classmethod
    def from_graphql(
        cls,
        graphql_qa_pair: ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairs,
    ) -> "QuestionAnswerPair":
        metadata = None
        if graphql_qa_pair.typed_metadata:
            metadata = Metadata(
                in_tokens=graphql_qa_pair.typed_metadata.in_tokens,
                out_tokens=graphql_qa_pair.typed_metadata.out_tokens,
                duration_seconds=graphql_qa_pair.typed_metadata.duration_seconds,
            )

        return cls(
            id=graphql_qa_pair.id,
            input_under_test=graphql_qa_pair.input_under_test,
            llm_output=graphql_qa_pair.llm_output,
            file_ids=graphql_qa_pair.typed_file_ids,
            context=graphql_qa_pair.context or {},
            output_context=graphql_qa_pair.output_context or {},
            metadata=metadata,
            test_id=graphql_qa_pair.test.test_id if graphql_qa_pair.test else None,
            local_evals=(
                [
                    LocalEvalUploadInputType(
                        question_answer_pair_id=graphql_qa_pair.id,
                        score=eval.score,
                        feedback=eval.feedback,
                        name="local_eval",
                    )
                    for eval in graphql_qa_pair.local_evals
                ]
                if graphql_qa_pair.local_evals
                else []
            ),
        )

    def to_graphql(self) -> QuestionAnswerPairInputType:
        return QuestionAnswerPairInputType(
            input_under_test=self.input_under_test,
            file_ids=self.file_ids,
            context=self.context,
            output_context=self.output_context,
            llm_output=self.llm_output,
            metadata=(
                MetadataType(**self.metadata.model_dump()) if self.metadata else None
            ),
            test_id=self.test_id,
        )


class OperatorInput(BaseModel):
    input: str
    model_output: str
    context: dict[str, Any] | None = None
    output_context: dict[str, Any] | None = None
    files: dict[str, BytesIO] | None = None

    model_config = {
        "arbitrary_types_allowed": True,
        "protected_namespaces": (),
    }


class OperatorOutput(BaseModel):
    name: str
    score: float
    explanation: str


ModelCustomOperatorFunctionType = Callable[[OperatorInput], OperatorOutput]


class CustomModelInput(BaseModel):
    input_under_test: str
    context: dict[str, Any]
    files: dict[str, BytesIO]

    model_config = {"arbitrary_types_allowed": True}


class CustomModelOutput(BaseModel):
    model_output: str | dict[str, Any] | QuestionAnswerPair

    model_config = {
        "arbitrary_types_allowed": True,
        "protected_namespaces": (),
    }


SimpleModelFunctionType = Callable[[str], str]

ModelFunctionWithFilesAndContextType = Callable[[CustomModelInput], CustomModelOutput]

ModelFunctionType = SimpleModelFunctionType | ModelFunctionWithFilesAndContextType



import click
import sys


def prompt_user_for_rag_suite():
    suites = list_rag_suites()
    click.echo("Rag Suites:")
    click.echo(
        "\n".join([f"{i}: {s['id']} {s['query']}" for i, s in enumerate(suites)])
    )

    idx = click.prompt("Enter the number of the rag suite to run", type=int)
    while not 0 <= idx <= len(suites):
        idx = click.prompt("Invalid choice. Retry", type=int)
    suite_id = suites[idx]["id"]
    return suite_id


def display_error_and_exit(message: str):
    click.secho(message, fg="red")
    sys.exit(1)



import asyncio
import concurrent.futures._base
import inspect
import json
import os
from time import time
from typing import Any, Callable, cast, overload

import requests
import vals.sdk.patch as patch
from pydantic import BaseModel, PrivateAttr
from tqdm import tqdm
from vals.graphql_client.client import UNSET, Client
from vals.graphql_client.get_operators import GetOperatorsOperators
from vals.graphql_client.input_types import (
    LocalEvalUploadInputType,
    MetadataType,
    ParameterInputType,
    QuestionAnswerPairInputType,
)
from vals.sdk.inspect_wrapper import InspectWrapper
from vals.sdk.run import Run
from vals.sdk.types import (
    Check,
    File,
    ModelCustomOperatorFunctionType,
    ModelFunctionType,
    ModelFunctionWithFilesAndContextType,
    OperatorInput,
    QuestionAnswerPair,
    RunParameters,
    RunStatus,
    SimpleModelFunctionType,
    Test,
    TestSuiteMetadata,
)
from vals.sdk.util import (
    _get_auth_token,
    be_host,
    download_files_bulk,
    fe_host,
    get_ariadne_client,
    md5_hash,
    parse_file_id,
    read_file,
)


class Suite(BaseModel):
    id: str | None = None

    title: str
    description: str = ""
    global_checks: list[Check] = []
    tests: list[Test] = []

    """
    Helper to see if we've instantiated the suite, but haven't 
    fetched it from the server yet. 
    """
    _client: Client = PrivateAttr(default_factory=get_ariadne_client)

    @classmethod
    async def list_suites(
        cls, limit=50, offset=0, search=""
    ) -> list[TestSuiteMetadata]:
        """
        Generate a list of all the test suites on the server.

        limit: Total number to return
        offset: Start list at this index
        """
        client = get_ariadne_client()
        gql_response = await client.get_test_suites_with_count(
            limit=limit,
            offset=offset,
            search=search,
        )
        gql_suites = gql_response.test_suites_with_count.test_suites
        return [TestSuiteMetadata.from_graphql(gql_suite) for gql_suite in gql_suites]

    @classmethod
    async def from_id(
        cls,
        suite_id: str,
        download_files: bool = False,
        download_path: str | None = None,
        max_concurrent_downloads: int = 50,
    ) -> "Suite":
        """
        Create a new local test suite based on the data from the server.
        """
        client = get_ariadne_client()
        suite_query = await client.get_test_suite_data(suite_id)
        suite_data = suite_query.test_suite

        title = suite_data.title
        description = suite_data.description

        global_checks = []
        if suite_data.global_checks is not None:
            global_checks = [
                Check.from_graphql(check)
                for check in json.loads(suite_data.global_checks)
            ]

        tests = []
        offset = 0
        page_size = 200
        have_pulled_all_tests = False

        while not have_pulled_all_tests:
            test_data = await client.get_test_data(suite_id, offset, page_size)

            for graphql_test in test_data.tests_with_count.tests:
                test = Test.from_graphql_test(graphql_test)
                tests.append(test)

            if len(tests) >= test_data.tests_with_count.count:
                have_pulled_all_tests = True
            else:
                offset += page_size

        suite = cls(
            id=suite_id,
            title=title,
            description=description,
            global_checks=global_checks,
            tests=tests,
        )

        if download_files:
            file_ids = []
            path = suite.title if download_path is None else download_path

            for test in tests:
                file_ids.extend([file.file_id for file in test.files_under_test])

            if len(file_ids) > 0:
                name_to_path_map = await download_files_bulk(
                    file_ids, path, max_concurrent_downloads=max_concurrent_downloads
                )
                test.files_under_test = [
                    File(
                        file_name=file.file_name,
                        file_id=file.file_id,
                        path=name_to_path_map[file.file_name],
                    )
                    for file in test.files_under_test
                ]

        return suite

    @classmethod
    async def from_dict(cls, data: dict[str, Any]) -> "Suite":
        """
        Imports the test suite from a dictionary - useful if
        importing from the old format.

        Does not create the suite - you must call suite.create() after construction.
        """

        title = data["title"]
        description = data.get("description", "")
        global_checks = [
            Check.from_graphql(check_dict)
            for check_dict in data.get("global_checks", [])
        ]
        tests = [
            Test(
                input_under_test=test["input_under_test"],
                checks=[
                    Check.from_graphql(check_dict) for check_dict in test["checks"]
                ],
                golden_output=test.get("golden_output", ""),
                tags=test.get("tags", []),
                files_under_test=test.get("files_under_test", []),
                context=test.get("context", {}),
            )
            for test in data["tests"]
        ]
        suite = cls(
            title=title,
            description=description,
            global_checks=global_checks,
            tests=tests,
        )
        await suite._validate_suite()
        return suite

    @classmethod
    async def from_json_file(cls, file_path: str) -> "Suite":
        """
        Imports the test suite from a local JSON file.
        """
        with open(file_path, "r") as f:
            data = json.load(f)
        return await cls.from_dict(data)

    @classmethod
    async def from_inspect_json_file(
        cls, file_path: str, suite_title: str, suite_description: str
    ) -> "Suite":
        """
        Imports the test suite from a local JSON file.
        """
        with open(file_path, "r") as f:
            data = json.load(f)

        suite = {"title": suite_title, "description": suite_description, "tests": []}

        for test in data:
            inspect_context = {
                "target": test.get("target", ""),
                "task_state": {
                    k: v for k, v in test.items() if k != "target" and k != "input"
                },
            }

            files_under_test = test.get("metadata", {}).get("documents_to_upload", [])
            context = {"inspect_context": inspect_context}

            suite["tests"].append(
                {
                    "input_under_test": test["input"],
                    "checks": [],
                    "files_under_test": files_under_test,
                    "context": context,
                    "tags": test.get("metadata", {}).get("tags", []),
                }
            )

        return await cls.from_dict(suite)

    @property
    def url(self):
        return f"{fe_host()}/suites/{self.id}"

    def to_dict(self) -> dict[str, Any]:
        """
        Converts the test suite to a dictionary.
        """
        return self.model_dump(exclude_none=True, exclude_defaults=True)

    def to_json_file(self, file_path: str) -> None:
        """
        Converts the test suite to a JSON file.
        """
        with open(file_path, "w") as f:
            f.write(self.to_json_string())

    def to_csv_string(self) -> str:
        """
        Like to_csv_file, but returns the file as a string instead of writing it to a file.
        """
        if not self.id:
            raise Exception("Suite has not been created yet.")

        url = f"{be_host()}/export_tests_to_file/?suite_id={self.id}"
        response = requests.post(
            url,
            headers={"Authorization": _get_auth_token()},
        )
        if response.status_code != 200:
            raise Exception(f"Failed to export tests: {response.text}")

        return response.text

    def to_json_string(self) -> str:
        """
        Converts the test suite to a JSON string.
        """
        if not self.id:
            raise Exception("Suite has not been created yet.")

        url = f"{be_host()}/export_tests_to_json/?suite_id={self.id}"
        response = requests.post(
            url,
            headers={"Authorization": _get_auth_token()},
        )

        if response.status_code != 200:
            raise Exception(f"Failed to export tests: {response.text}")

        return response.text

    def to_csv_file(self, file_path: str) -> None:
        """
        Converts the test suite to a CSV file.
        """
        with open(file_path, "w") as f:
            f.write(self.to_csv_string())

    async def create(self, force_creation: bool = False) -> None:
        """
        Creates the test suite on the server.

        force_creation: If True, will create a new test suite, even if it
        already exists.
        """
        if force_creation:
            self.id = None

        if self.id is not None:
            raise Exception("This suite has already been created.")

        await self._validate_suite()

        # Create suite object on the server
        suite = await self._client.create_or_update_test_suite(
            "0", self.title, self.description
        )
        if suite.update_test_suite is None:
            raise Exception("Unable to update the test suite.")
        self.id = suite.update_test_suite.test_suite.id

        await self._upload_global_checks()
        await self._upload_files()
        await self._upload_tests(create_only=True)

    async def delete(self) -> None:
        """
        Deletes the test suite from the server.
        """
        if self.id is None:
            raise Exception(
                "This suite has not been created yet, so there's nothing to delete"
            )

        await self._client.delete_test_suite(self.id)

    async def update(self, upload_files_path: str | None = None) -> None:
        """
        Pushes any local changes to the the test suite object to the server.
        """
        if self.id is None:
            raise Exception(
                "This suite has not been created yet, so there's nothing to update"
            )

        await self._client.create_or_update_test_suite(
            self.id, self.title, self.description
        )

        path = self.title if upload_files_path is None else upload_files_path

        await self._upload_files(path)
        await self._upload_global_checks()
        await self._upload_tests(create_only=False)

        # Remove any tests that are no longer used after the update.
        await self._client.remove_old_tests(self.id, [test._id for test in self.tests])

    @overload
    async def run(
        self,
        model: str,
        model_name: str = "sdk",
        run_name: str | None = None,
        wait_for_completion: bool = False,
        parameters: RunParameters | None = None,
    ) -> Run:
        """
        Runs based on a model string (e.g. "gpt-4o").
        """
        pass

    @overload
    async def run(
        self,
        model: ModelFunctionType,
        model_name: str = "sdk",
        run_name: str | None = None,
        wait_for_completion: bool = False,
        parameters: RunParameters | None = None,
        upload_concurrency: int | None = None,
    ) -> Run:
        """
        Runs based on a model function. This can either be a simple model function, which just takes
        an input as a string and produces an output string, or a model function that takes in the input string,
        a dictionary of filename to file contents, and a context dictionary.

        Args:
            upload_concurrency: How frequently to upload QA pairs to the server. Defaults to parameters.parallelism.
        """
        pass

    @overload
    async def run(
        self,
        model: list[QuestionAnswerPair],
        model_name: str = "sdk",
        run_name: str | None = None,
        wait_for_completion: bool = False,
        parameters: RunParameters | None = None,
    ) -> Run:
        """
        Runs based on on a list of question-answer pairs, which contain inputs and outputs.

        IMPORTANT NOTE: The combined "inputs" of the question-answer pairs (the input under test, the context,
        and the files) need to match exactly the inputs of the tests in the suite, otherwise, Vals will not
        know how to match to the auto-eval correctly.
        """
        pass

    @overload
    async def run(
        self,
        model: InspectWrapper,
        model_name: str = "sdk",
        run_name: str | None = None,
        wait_for_completion: bool = False,
        parameters: RunParameters | None = None,
    ) -> Run:
        """
        Runs based on on an InspectWrapper object.
        """
        pass

    async def _safe_local_execution(self, coro, run_id: str | None = None):
        """Wrapper to handle interruptions during local execution"""
        try:
            return await coro
        except (
            KeyboardInterrupt,
            asyncio.CancelledError,
            concurrent.futures._base.CancelledError,
        ):
            if run_id:
                await self._client.update_run_status(
                    run_id=run_id, status=RunStatus.ERROR
                )
            raise
        except Exception as e:
            if run_id:
                await self._client.update_run_status(
                    run_id=run_id, status=RunStatus.ERROR
                )
            raise

    async def run(
        self,
        model: str | ModelFunctionType | list[QuestionAnswerPair] | InspectWrapper,
        model_name: str = "sdk",
        run_name: str | None = None,
        wait_for_completion: bool = False,
        parameters: RunParameters | None = None,
        upload_concurrency: int | None = None,
        custom_operators: list[ModelCustomOperatorFunctionType] | None = None,
        eval_model_name: str | None = None,
        run_id: str | None = None,
        qa_set_id: str | None = None,
        remaining_tests: list[Test] | None = None,
        uploaded_qa_pairs: list[QuestionAnswerPairInputType] | None = None,
    ) -> Run:
        """
        Base method for running the test suite. See overloads for documentation.

        Args:
            upload_concurrency: How frequently to upload QA pairs to the server.
                         Defaults to parameters.parallelism.
            custom_operator: A custom operator function that takes in an OperatorInput and returns an OperatorOutput.
        """
        if self.id is None:
            raise Exception(
                "This suite has not been created yet. Call suite.create() before calling suite.run()"
            )
        if parameters is None:
            parameters = RunParameters()

        if custom_operators is None:
            custom_operators = []

        if eval_model_name is not None:
            parameters.eval_model = eval_model_name

        if upload_concurrency is None:
            upload_concurrency = parameters.parallelism

        if uploaded_qa_pairs is None:
            uploaded_qa_pairs = []

        if isinstance(model, str) and len(custom_operators) > 0:
            raise Exception(
                "Custom operator functions are not supported when using a model string (e.g. a nonlocal model)."
            )

        if isinstance(model, InspectWrapper):
            model_name = model.model_name
            eval_model_name = model.eval_model_name
            inspect_wrapper = model
            custom_operators = model.get_custom_operators()
            model = model.get_custom_model()

        parameter_json = parameters.model_dump()
        del parameter_json[
            "parallelism"
        ]  # Need to explicitly map parallelism to maximum_threads
        parameter_input = ParameterInputType(
            **parameter_json,
            model_under_test="",
            maximum_threads=parameters.parallelism,
        )

        # Collate the local output pairs if we're using a model function.
        if isinstance(model, Callable) or isinstance(model, InspectWrapper):
            # Generate the QA pairs from the model function
            parameter_input.model_under_test = model_name
            if run_id is None and qa_set_id is None:
                qa_set_id, run_id = await self._create_empty_qa_set(
                    parameter_input.model_dump(), model_name, run_name
                )
                print(f"Created run with run id: {run_id}")
            elif (
                run_id is None
                and qa_set_id is not None
                or run_id is not None
                and qa_set_id is None
            ):
                raise Exception("Must provide both run_id and qa_set_id or neither.")

            # Wrap the QA pair generation with safe execution
            uploaded_qa_pairs += await self._safe_local_execution(
                self._generate_qa_pairs_from_function(
                    model,
                    parameter_input,
                    qa_set_id,
                    upload_concurrency,
                    remaining_tests,
                ),
                run_id,
            )
        elif isinstance(model, list):
            # Use the QA pairs we are already provided

            parameter_input.model_under_test = model_name
            qa_pairs = model
            uploadable_qa_pairs = [qa_pair.to_graphql() for qa_pair in qa_pairs]
            qa_set_id, run_id = await self._create_empty_qa_set(
                parameter_input.model_dump(), model_name, run_name
            )
            uploaded_qa_pairs = []
            for i in range(0, len(uploadable_qa_pairs), upload_concurrency):
                batch_to_upload = uploadable_qa_pairs[i : i + upload_concurrency]
                response = await self._client.batch_add_question_answer_pairs(
                    qa_set_id, batch_to_upload
                )
                uploaded_qa_pairs.extend(
                    [
                        await self._deserialize_qa_pair_file_ids(qa_pair)
                        for qa_pair in response.batch_add_question_answer_pairs.question_answer_pairs
                    ]
                )

        elif isinstance(model, str):
            # Just use a model string (e.g. "gpt-4o")
            parameter_input.model_under_test = model
            qa_set_id = None
        else:
            raise Exception(f"Got unexpected type for model: {type(model)}")

        # Start and pull the run.
        if len(custom_operators) > 0:
            # Wrap the local eval with safe execution
            await self._safe_local_execution(
                self._upload_local_eval(
                    custom_operators,
                    parameter_input,
                    qa_set_id,
                    uploaded_qa_pairs,
                    upload_concurrency,
                ),
                run_id,
            )

        response = await self._client.start_run(
            self.id,
            parameter_input,
            qa_set_id or UNSET,
            run_name or UNSET,
            run_id or UNSET,
        )

        if response.start_run is None:
            raise Exception("Unable to start the run.")

        run_id = response.start_run.run_id

        run = await Run.from_id(run_id)

        if wait_for_completion:
            await run.wait_for_run_completion()

        await run.pull()

        return run

    async def _upload_files(self, upload_files_path: str | None = None):
        """
        Helper method to upload the files to the server.
        Uploads any tests.files_under_test that haven't been uploaded yet.
        """
        if self.id is None:
            raise Exception("This suite has not been created yet.")

        for test in self.tests:
            # Convert files_under_test to a list of File objects if they're not already
            normalized_files = []
            for file_item in test.files_under_test:
                if isinstance(file_item, str):
                    # Handle string path
                    normalized_files.append(
                        File(file_name=os.path.basename(file_item), path=file_item)
                    )
                elif isinstance(file_item, dict):
                    # Handle dictionary representation
                    file_name = file_item.get("file_name") or os.path.basename(
                        file_item.get("path", "")
                    )
                    normalized_files.append(
                        File(
                            file_name=file_name,
                            file_id=file_item.get("file_id"),
                            path=file_item.get("path"),
                            hash=file_item.get("hash"),
                        )
                    )
                elif isinstance(file_item, File):
                    # Already a File object
                    normalized_files.append(file_item)
                else:
                    raise ValueError(f"Unexpected file type: {type(file_item)}")

            # Replace the original files_under_test with normalized File objects
            test.files_under_test = normalized_files

            # Process each file
            for file in test.files_under_test:
                # Determine the file path
                file_path = None

                # Case 1: File has a path
                if file.path is not None and os.path.exists(file.path):
                    file_path = file.path
                # Case 2: File has no path but we have upload_files_path
                elif upload_files_path is not None:
                    # Try standard path
                    potential_path = os.path.join(upload_files_path, file.file_name)
                    if os.path.exists(potential_path):
                        file_path = potential_path
                    # Try hash subdirectory path if hash exists
                    elif file.hash is not None:
                        hash_path = os.path.join(
                            upload_files_path, file.hash, file.file_name
                        )
                        if os.path.exists(hash_path):
                            file_path = hash_path

                # Skip if we couldn't find the file
                if file_path is None:
                    continue

                # Calculate file hash
                with open(file_path, "rb") as f:
                    file_hash = md5_hash(f)

                # Case: File is new or hash has changed
                if file.file_id is None or file.hash != file_hash:
                    # Check if this hash already exists in another file in the test
                    if file_hash in [
                        f.hash
                        for f in test.files_under_test
                        if f != file and f.hash is not None
                    ]:
                        print(
                            f"File {file.file_name} with same hash already exists in the test suite."
                        )
                        continue

                    file.file_id = self._upload_file(self.id, file_path)
                    file.hash = file_hash

            # Update file IDs for the test
            test._file_ids = [
                file.file_id
                for file in test.files_under_test
                if file.file_id is not None
            ]

    async def _upload_tests(self, create_only: bool = True) -> None:
        """
        Helper method to upload the tests to the server in batches of 100.
        """
        if self.id is None:
            raise Exception("This suite has not been created yet.")

        # Upload tests in batches of 100
        created_tests = []
        test_mutations = [test.to_test_mutation_info(self.id) for test in self.tests]

        with tqdm(total=len(test_mutations), desc="Uploading tests") as pbar:
            for i in range(0, len(test_mutations), 25):
                batch = test_mutations[i : i + 25]
                batch_result = await self._client.add_batch_tests(
                    tests=batch,
                    create_only=create_only,
                )
                if batch_result.batch_update_test is None:
                    raise Exception("Unable to add tests to the test suite.")
                created_tests.extend(batch_result.batch_update_test.tests)
                pbar.update(len(batch))

        # Update the local suite with the new tests to ensure everything is in sync.
        self.tests = [Test.from_graphql_test(test) for test in created_tests]

    async def _validate_checks(
        self, check: Check, operators_dict: dict[str, GetOperatorsOperators]
    ) -> None:
        """Helper method to ensure that operator is correct"""
        if check.operator not in operators_dict:
            raise ValueError(f"Invalid operator: {check.operator}")
        operator = operators_dict[check.operator]
        if not operator.is_unary and (check.criteria is None or check.criteria == ""):
            raise ValueError(
                f"Operator {operator.name_in_doc} is unary, but no criteria was provided."
            )

    async def _validate_suite(self) -> None:
        """Helper method to ensure that all operator strings are correct. Most of the validation is done by Pydantic
        but this handles things Pydantic can't check for, like whether a file exists."""
        operators = (await self._client.get_operators()).operators
        base_operators_dict = {op.name_in_doc: op for op in operators}
        custom_operators = (
            await self._client.get_active_custom_operators(0, 200)
        ).custom_operators.operators
        custom_operators_dict = {op.name: op for op in custom_operators}
        operators_dict = {**base_operators_dict, **custom_operators_dict}

        for test in self.tests:
            for check in test.checks:
                await self._validate_checks(check, operators_dict)

            for file in test.files_under_test:
                if isinstance(file, str):
                    file_path = file
                elif isinstance(file, dict):
                    file_path = file["path"]
                elif isinstance(file, File):
                    file_path = file.path
                else:
                    raise ValueError(f"Unexpected file type: {type(file)}")
                if not os.path.exists(file_path):
                    raise ValueError(f"File does not exist: {file_path}")
                if not os.path.isfile(file_path):
                    raise ValueError(f"Path is a directory: {file_path}")

        for check in self.global_checks:
            await self._validate_checks(check, operators_dict)

    async def _upload_global_checks(self) -> None:
        """
        Helper method to upload the global checks to the server.
        """
        if self.id is None:
            raise Exception("This suite has not been created yet.")

        await self._client.update_global_checks(
            self.id,
            [gc.to_graphql_input() for gc in self.global_checks],
        )

    async def _create_empty_qa_set(
        self,
        parameters: dict[str, int | float | str | bool] = {},
        model_under_test: str | None = None,
        run_name: str | None = None,
    ) -> str:
        """Creates an empty QA set and returns its ID."""
        if self.id is None:
            raise Exception(
                "This suite has not been created yet, so we can't create a QA set from it."
            )

        response = await self._client.create_question_answer_set(
            self.id,
            [],
            parameters,
            model_under_test or "sdk",
            run_name,
        )

        if response.create_question_answer_set is None:
            raise Exception("Unable to create the question-answer set.")

        return (
            response.create_question_answer_set.question_answer_set.id,
            response.create_question_answer_set.run_id,
        )

    async def _upload_local_eval(
        self,
        custom_operators: list[ModelCustomOperatorFunctionType],
        parameter_input: ParameterInputType,
        qa_set_id: str,
        qa_pairs: list[QuestionAnswerPairInputType],
        upload_concurrency: int,
    ) -> str:
        """Process QA pairs through custom operator and upload results in batches."""
        if self.id is None:
            raise Exception("This suite has not been created yet.")

        # Create a semaphore to limit concurrent processing
        semaphore = asyncio.Semaphore(parameter_input.maximum_threads)
        operator_results = []
        last_uploaded_idx = 0
        upload_tasks = []

        async def process_and_upload_eval(qa_pair):
            async with semaphore:
                # Process through all custom operators
                results = await asyncio.gather(
                    *(
                        self._process_single_local_eval(custom_operator, qa_pair)
                        for custom_operator in custom_operators
                    )
                )
                operator_results.extend(results)
                pbar.update(1)

                # Upload batch when we reach batch_size
                nonlocal last_uploaded_idx
                while len(operator_results) >= last_uploaded_idx + upload_concurrency:
                    pbar.set_description(
                        f"Processing operator evaluations (uploading evaluations {last_uploaded_idx} to {last_uploaded_idx + upload_concurrency})"
                    )
                    batch_to_upload = operator_results[
                        last_uploaded_idx : last_uploaded_idx + upload_concurrency
                    ]
                    task = asyncio.create_task(
                        self._client.upload_local_evaluation(qa_set_id, batch_to_upload)
                    )
                    upload_tasks.append(task)
                    last_uploaded_idx += upload_concurrency

        # Process all QA pairs concurrently with limited concurrency
        with tqdm(total=len(qa_pairs), desc="Processing operator evaluations") as pbar:
            await asyncio.gather(*(process_and_upload_eval(qa) for qa in qa_pairs))

        # Upload any remaining results
        if last_uploaded_idx < len(operator_results):
            batch_to_upload = operator_results[last_uploaded_idx:]
            task = asyncio.create_task(
                self._client.upload_local_evaluation(qa_set_id, batch_to_upload)
            )
            upload_tasks.append(task)

        # Wait for all uploads to complete
        if upload_tasks:
            await asyncio.gather(*upload_tasks)

    async def _process_single_test(
        self,
        test: Test,
        model_function: ModelFunctionType,
        is_simple_model_function: bool,
    ) -> QuestionAnswerPairInputType:
        """Inner implementation of process_single_test"""
        files = {}
        for file_id in test._file_ids:
            _, file_name, _ = parse_file_id(file_id)
            files[file_name] = read_file(file_id)

        time_start = time()
        in_tokens_start = patch.in_tokens
        out_tokens_start = patch.out_tokens

        # Get the current event loop
        loop = asyncio.get_running_loop()

        if is_simple_model_function:
            casted_model_function = cast(SimpleModelFunctionType, model_function)
            if inspect.iscoroutinefunction(casted_model_function):
                # Already async, just await it
                output = await casted_model_function(test.input_under_test)
            else:
                # Run synchronous function in a thread pool to avoid blocking
                output = await loop.run_in_executor(
                    None, lambda: casted_model_function(test.input_under_test)
                )
        else:
            casted_model_function = cast(
                ModelFunctionWithFilesAndContextType, model_function
            )
            if inspect.iscoroutinefunction(casted_model_function):
                # Already async, just await it
                output = await casted_model_function(
                    test.input_under_test, files, test.context
                )
            else:
                # Run synchronous function in a thread pool to avoid blocking
                output = await loop.run_in_executor(
                    None,
                    lambda: casted_model_function(
                        test.input_under_test, files, test.context
                    ),
                )

        time_end = time()
        in_tokens_end = patch.in_tokens
        out_tokens_end = patch.out_tokens

        return await self._process_model_output(
            output,
            test,
            test._file_ids,
            time_start,
            time_end,
            in_tokens_start,
            out_tokens_start,
            in_tokens_end,
            out_tokens_end,
        )

    async def _process_single_local_eval(
        self,
        custom_operator: ModelCustomOperatorFunctionType,
        qa_pair: QuestionAnswerPairInputType,
    ) -> LocalEvalUploadInputType:
        """Inner implementation of process_single_local_eval"""
        files = {}
        for file_id in qa_pair.file_ids:
            _, file_name, _ = parse_file_id(file_id)
            files[file_name] = read_file(file_id)

        operator_input = OperatorInput(
            input=qa_pair.input_under_test,
            model_output=qa_pair.llm_output,
            files=files,
            context=qa_pair.context,
            output_context=qa_pair.output_context,
        )

        # Get the current event loop
        loop = asyncio.get_running_loop()

        if inspect.iscoroutinefunction(custom_operator):
            # Already async, just await it
            result = await custom_operator(operator_input)
        else:
            # Run synchronous function in a thread pool to avoid blocking
            result = await loop.run_in_executor(
                None, lambda: custom_operator(operator_input)
            )

        return LocalEvalUploadInputType(
            questionAnswerPairId=qa_pair.id,
            score=result.score,
            feedback=result.explanation,
            name=result.name,
        )

    async def _deserialize_qa_pair_file_ids(self, qa_pair):
        """Helper method to deserialize file_ids in a QA pair if they're stored as a string."""
        if hasattr(qa_pair, "file_ids") and isinstance(qa_pair.file_ids, str):
            qa_pair.file_ids = json.loads(qa_pair.file_ids)
        return qa_pair

    async def _generate_qa_pairs_from_function(
        self,
        model_function: ModelFunctionType,
        parameter_input: ParameterInputType,
        qa_set_id: str,
        upload_concurrency: int,
        remaining_tests: list[Test] | None = None,
    ) -> list[QuestionAnswerPairInputType]:
        if self.id is None:
            raise Exception(
                "This suite has not been created yet, so we can't create a QA set from it."
            )
        # Pull latest suite data to ensure we're up to date
        updated_suite = await Suite.from_id(self.id)
        for field in self.__fields__:
            setattr(self, field, getattr(updated_suite, field))

        # Inspect the model function
        sig = inspect.signature(model_function)
        num_params = len(sig.parameters)
        if num_params not in [1, 3]:
            raise ValueError("Model function must accept either 1 or 3 parameters")
        is_simple_model_function = num_params == 1

        # Create a semaphore to limit concurrent processing
        semaphore = asyncio.Semaphore(parameter_input.maximum_threads)
        last_uploaded_idx = 0
        upload_tasks = []
        qa_pairs: list[QuestionAnswerPairInputType] = []

        async def process_and_upload_test(test):
            async with semaphore:
                result = await self._process_single_test(
                    test, model_function, is_simple_model_function
                )
                qa_pairs.append(result)
                pbar.update(1)

                # Upload batch when we reach batch_size
                nonlocal last_uploaded_idx
                while len(qa_pairs) >= last_uploaded_idx + upload_concurrency:
                    pbar.set_description(
                        f"Processing tests (uploading outputs {last_uploaded_idx} to {last_uploaded_idx + upload_concurrency})"
                    )
                    batch_to_upload = qa_pairs[
                        last_uploaded_idx : last_uploaded_idx + upload_concurrency
                    ]
                    task = asyncio.create_task(
                        self._client.batch_add_question_answer_pairs(
                            qa_set_id, batch_to_upload
                        )
                    )
                    upload_tasks.append(task)
                    last_uploaded_idx += upload_concurrency

        # Process all tests concurrently with limited concurrency

        with tqdm(
            total=(
                len(self.tests) if remaining_tests is None else len(remaining_tests)
            ),
            desc="Processing tests",
        ) as pbar:
            await asyncio.gather(
                *(
                    process_and_upload_test(test)
                    for test in (
                        self.tests if remaining_tests is None else remaining_tests
                    )
                )
            )

        # Upload any remaining pairs
        if last_uploaded_idx < len(qa_pairs):
            remaining_pairs = qa_pairs[last_uploaded_idx:]
            task = asyncio.create_task(
                self._client.batch_add_question_answer_pairs(qa_set_id, remaining_pairs)
            )
            upload_tasks.append(task)

        # Wait for all uploads to complete
        uploaded_qa_pairs = []
        if upload_tasks:
            uploaded_qa_pairs.extend(
                [
                    await self._deserialize_qa_pair_file_ids(qa_pair)
                    for result in await asyncio.gather(*upload_tasks)
                    for qa_pair in result.batch_add_question_answer_pairs.question_answer_pairs
                ]
            )

        await self._client.mark_question_answer_set_as_complete(qa_set_id)

        return uploaded_qa_pairs

    def _upload_file(self, suite_id: str, file_path: str) -> str:
        with open(file_path, "rb") as f:
            response = requests.post(
                f"{be_host()}/upload_file/?test_suite_id={suite_id}",
                files={"file": f},
                headers={"Authorization": _get_auth_token()},
            )
            if response.status_code != 200:
                raise Exception(f"Failed to upload file {file_path}")
            return response.json()["file_id"]

    async def _process_model_output(
        self,
        output: str | dict | QuestionAnswerPairInputType,
        test: Test,
        file_ids: list[str],
        time_start: float,
        time_end: float,
        in_tokens_start: int,
        out_tokens_start: int,
        in_tokens_end: int,
        out_tokens_end: int,
    ) -> QuestionAnswerPairInputType:
        """Helper function to process model output into a QuestionAnswerPairInputType.
        The output can now be either a string, a dict or a QuestionAnswerPairInputType, so we need to handle all cases.
        """
        if isinstance(output, QuestionAnswerPairInputType):
            return output

        # If output is just a string, treat it as llm_output
        if isinstance(output, str):
            return QuestionAnswerPairInputType(
                input_under_test=test.input_under_test,
                file_ids=file_ids,
                context=test.context,
                llm_output=output,
                metadata=MetadataType(
                    in_tokens=in_tokens_end - in_tokens_start,
                    out_tokens=out_tokens_end - out_tokens_start,
                    duration_seconds=time_end - time_start,
                ),
                test_id=test._id,
            )

        # If output is a dict, use provided values or defaults
        llm_output = output.get("llm_output", "")
        metadata = output.get("metadata", {})
        in_tokens = metadata.get("in_tokens", in_tokens_end - in_tokens_start)
        out_tokens = metadata.get("out_tokens", out_tokens_end - out_tokens_start)
        output_context = output.get("output_context", None)

        return QuestionAnswerPairInputType(
            input_under_test=test.input_under_test,
            file_ids=file_ids,
            context=test.context,
            llm_output=llm_output,
            output_context=output_context,
            metadata=MetadataType(
                in_tokens=in_tokens,
                out_tokens=out_tokens,
                duration_seconds=time_end - time_start,
            ),
            test_id=test._id,
        )



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import Optional

from pydantic import Field

from .base_model import BaseModel


class CreateOrUpdateTestSuite(BaseModel):
    update_test_suite: Optional["CreateOrUpdateTestSuiteUpdateTestSuite"] = Field(
        alias="updateTestSuite"
    )


class CreateOrUpdateTestSuiteUpdateTestSuite(BaseModel):
    test_suite: "CreateOrUpdateTestSuiteUpdateTestSuiteTestSuite" = Field(
        alias="testSuite"
    )


class CreateOrUpdateTestSuiteUpdateTestSuiteTestSuite(BaseModel):
    description: str
    id: str
    org: str
    title: str


CreateOrUpdateTestSuite.model_rebuild()
CreateOrUpdateTestSuiteUpdateTestSuite.model_rebuild()



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import Optional

from pydantic import Field

from .base_model import BaseModel


class RemoveOldTests(BaseModel):
    remove_unused_tests: Optional["RemoveOldTestsRemoveUnusedTests"] = Field(
        alias="removeUnusedTests"
    )


class RemoveOldTestsRemoveUnusedTests(BaseModel):
    success: bool


RemoveOldTests.model_rebuild()



from setuptools import find_packages, setup

setup(
    name="valsai",
    author="Langston Nashold, Rayan Krishnan",
    packages=find_packages(),
    include_package_data=True,
    package_data={"": ["jsonschemas/*"]},
    install_requires=[
        "Click",
        "gql",
        "tqdm",
        "requests",
        "pandas",
        "descope",
        "pydantic",
        "httpx",
        "tabulate",
        "inspect-ai",
    ],
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    extras_require={
        # Requirements only needed for development, not for users
        "dev": ["ariadne-codegen"],
        # Requirements for parsing utilities we provide
        "parsing": ["pypdf", "pypandoc-binary"],
    },
    entry_points={
        "console_scripts": [
            "vals = vals.cli.main:cli",
        ],
    },
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
)



# Generated by ariadne-codegen
# Source: vals/graphql/

from datetime import datetime
from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel


class ListQuestionAnswerPairs(BaseModel):
    question_answer_pairs_with_count: (
        "ListQuestionAnswerPairsQuestionAnswerPairsWithCount"
    ) = Field(alias="questionAnswerPairsWithCount")


class ListQuestionAnswerPairsQuestionAnswerPairsWithCount(BaseModel):
    question_answer_pairs: List[
        "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairs"
    ] = Field(alias="questionAnswerPairs")
    count: int


class ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairs(BaseModel):
    id: Any
    input_under_test: str = Field(alias="inputUnderTest")
    llm_output: str = Field(alias="llmOutput")
    context: Any
    output_context: Any = Field(alias="outputContext")
    typed_metadata: (
        "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTypedMetadata"
    ) = Field(alias="typedMetadata")
    typed_file_ids: List[str] = Field(alias="typedFileIds")
    local_evals: List[
        "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsLocalEvals"
    ] = Field(alias="localEvals")
    test: Optional[
        "ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTest"
    ]


class ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTypedMetadata(
    BaseModel
):
    in_tokens: int = Field(alias="inTokens")
    out_tokens: int = Field(alias="outTokens")
    duration_seconds: float = Field(alias="durationSeconds")


class ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsLocalEvals(
    BaseModel
):
    id: str
    score: float
    feedback: str
    created_at: datetime = Field(alias="createdAt")


class ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairsTest(
    BaseModel
):
    test_id: str = Field(alias="testId")


ListQuestionAnswerPairs.model_rebuild()
ListQuestionAnswerPairsQuestionAnswerPairsWithCount.model_rebuild()
ListQuestionAnswerPairsQuestionAnswerPairsWithCountQuestionAnswerPairs.model_rebuild()



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import Optional

from pydantic import Field

from .base_model import BaseModel


class CreateRagSuite(BaseModel):
    update_rag_suite: Optional["CreateRagSuiteUpdateRagSuite"] = Field(
        alias="updateRagSuite"
    )


class CreateRagSuiteUpdateRagSuite(BaseModel):
    rag_suite: Optional["CreateRagSuiteUpdateRagSuiteRagSuite"] = Field(
        alias="ragSuite"
    )


class CreateRagSuiteUpdateRagSuiteRagSuite(BaseModel):
    id: str
    query: str


CreateRagSuite.model_rebuild()
CreateRagSuiteUpdateRagSuite.model_rebuild()



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import List

from pydantic import Field

from .base_model import BaseModel


class GetRagSuites(BaseModel):
    rag_suites: List["GetRagSuitesRagSuites"] = Field(alias="ragSuites")


class GetRagSuitesRagSuites(BaseModel):
    id: str
    org: str
    path: str
    query: str


GetRagSuites.model_rebuild()



# Generated by ariadne-codegen
# Source: vals/graphql/

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel


class PullTestResultsWithCount(BaseModel):
    test_results_with_count: "PullTestResultsWithCountTestResultsWithCount" = Field(
        alias="testResultsWithCount"
    )


class PullTestResultsWithCountTestResultsWithCount(BaseModel):
    test_results: List["PullTestResultsWithCountTestResultsWithCountTestResults"] = (
        Field(alias="testResults")
    )
    count: int


class PullTestResultsWithCountTestResultsWithCountTestResults(BaseModel):
    id: str
    llm_output: str = Field(alias="llmOutput")
    pass_percentage: float = Field(alias="passPercentage")
    pass_percentage_with_optional: float = Field(alias="passPercentageWithOptional")
    result_json: Any = Field(alias="resultJson")
    qa_pair: Optional[
        "PullTestResultsWithCountTestResultsWithCountTestResultsQaPair"
    ] = Field(alias="qaPair")
    test: "PullTestResultsWithCountTestResultsWithCountTestResultsTest"
    metadata: Optional[Any]


class PullTestResultsWithCountTestResultsWithCountTestResultsQaPair(BaseModel):
    context: Any
    output_context: Any = Field(alias="outputContext")
    error_message: str = Field(alias="errorMessage")


class PullTestResultsWithCountTestResultsWithCountTestResultsTest(BaseModel):
    test_id: str = Field(alias="testId")
    input_under_test: str = Field(alias="inputUnderTest")
    context: Any


PullTestResultsWithCount.model_rebuild()
PullTestResultsWithCountTestResultsWithCount.model_rebuild()
PullTestResultsWithCountTestResultsWithCountTestResults.model_rebuild()
